package core;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.awt.Point;


public abstract class Calc implements Misc
{

    public static long orientation(Point a, Point b)
    {
        if (b.y == a.y)
        {
            return 0;
        }
        if (b.x == a.x)
        {
            return 90 * Const.M;
        }

        long ang = Math.round(Const.M * (180 - Math.atan2(b.y - a.y, b.x - a.x) / Const.RAD));
        return ang < 0 ? 180 * Const.M + ang : ang % (180 * Const.M);
    }

    public static long distanceM(Point a, Point b)
    {
        return Math.round(Const.M * Math.sqrt((a.x - b.x) * (a.x - b.x) + (a.y - b.y) * (a.y - b.y)));
    }

    /**
     * Mask of points (x,y) that satisfy x^2 + y^2 < ceil(0.25 * (d^2 + d))
     * which is equivalent to x^2 + y^2 < 0.25 * (d^2 + d).
     */
    public static boolean[][] circleMask(int diameter)
    {
        int range = diameter / 2;
        boolean[][] mask = new boolean[2 * range + 1][2 * range + 1];
        for (int x = 0; x <= 2 * range; x++)
        {
            for (int y = 0; y <= 2 * range; y++)
            {
                mask[x][y] = (x - range) * (x - range) + (y - range) * (y - range) <
                             (diameter * diameter + diameter + 3) / 4;
            }
        }
        return mask;
    }

    public static double[][] gaussianMask(int range, double phi, double sigma_x, double sigma_y)
    {
        double sum = 0.0;

        double[] rotated_variance =
        {
            Math.cos(phi) * Math.cos(phi) / (sigma_x * sigma_x)
          + Math.sin(phi) * Math.sin(phi) / (sigma_y * sigma_y),

            Math.sin(phi) * Math.cos(phi) * (1 / (sigma_x * sigma_x) - 1 / (sigma_y * sigma_y)),

            Math.sin(phi) * Math.sin(phi) / (sigma_x * sigma_x)
          + Math.cos(phi) * Math.cos(phi) / (sigma_y * sigma_y)
        };

        double[][] mask = new double[2*range+1][2*range+1];
        for(int dx=-range; dx<=range; dx++)
        {
            for(int dy=-range; dy<=range; dy++)
            {
                double gauss = Math.exp(-(dx * dx * rotated_variance[0]
                                          + 2 * dx * dy * rotated_variance[1]
                                          + dy * dy * rotated_variance[2]) / 2.0);
                mask[range+dx][range+dy] = gauss;
                sum += gauss;
            }
        }

        for(int dx=-range; dx<=range; dx++)
        {
            for(int dy=-range; dy<=range; dy++)
            {
                mask[range+dx][range+dy] /= sum;
            }
        }
        return mask;
    }

    public static double[][] gaussianMask(int range, double sigma)
    {
        return gaussianMask(range, 0, sigma, sigma);
    }

    /**
     * Return a one-dimensional Gaussian mask where the sum of all
     * entries is 1.
     */
    public static double[] gaussian1d(int range, double sigma)
    {
        double sum = 0.0;
        final double factor = -1.0 / (2.0 * sigma * sigma);

        double[] mask = new double[2*range+1];
        for(int dx=-range; dx<=range; dx++)
        {
            double gauss = Math.exp(factor * (dx * dx));
            mask[range + dx] = gauss;
            sum += gauss;
        }

        for(int dx=-range; dx<=range; dx++)
        {
            mask[range + dx] /= sum;
        }
        return mask;
    }

    /**
     * Returns a one-dimensional Gaussian mask whose center is scaled
     * to the given value "scale".
     */
    public static double[] gaussian1d(final int range, final double sigma,
                                      final double scale)
    {
        final double factor = -1.0 / (2.0 * sigma * sigma);

        double[] mask = new double[2*range+1];
        for(int dx=-range; dx<=range; dx++)
        {
            mask[range + dx] = Math.exp(factor * (dx * dx)) * scale;
        }
        return mask;
    }

    /**
     * Return a one-dimensional mask of minus the derivative of a Gaussian
     * where the sum of all entries of the original gaussian is 1. The
     * inverse sign is chosen such that applying this mask to an array it
     * acts as the derivative of the array smeared by a Gaussian.
     */
    public static double[] gaussianDerivative(final int range, final double sigma)
    {
        double sum = 0.0;
        final double factor = -1.0 / (2.0 * sigma * sigma);

        double[] mask = new double[2*range+1];
        for(int dx=-range; dx<=range; dx++)
        {
            double gauss = Math.exp(factor * (dx * dx));
            mask[range + dx] = - 2.0 * factor * dx * gauss;
            sum += gauss;
        }

        for(int dx=-range; dx<=range; dx++)
        {
            mask[range + dx] /= sum;
        }
        return mask;
    }

    public static double[] gaussianDerivative2(final int range, final double sigma)
    {
        double sum = 0.0;
        final double factor = -1.0 / (sigma * sigma);
        final double factor2 = factor * factor;

        double[] mask = new double[2*range+1];
        for(int dx=-range; dx<=range; dx++)
        {
            double gauss = Math.exp(0.5 * factor * (dx * dx));
            mask[range + dx] = (factor + factor2 * dx * dx) * gauss;
            sum += gauss;
        }

        for(int dx=-range; dx<=range; dx++)
        {
            mask[range + dx] /= sum;
        }
        return mask;
    }

    public static long pow(int base, int exp)
    {
        long out = 1;
        for (int i = 0; i < exp; i++)
        {
            out *= base;
        }
        return out;
    }

    /**
     * Invert a monotonic function. If the input function is not monotonic
     * in the suitable sense, the result is undefined. 
     * @param function  A monotonically growing function.
     * @param guess     A guess for the values of the inverse function.
     * @param cutoff    The maximum result of the inverse function.
     * @param growing   This states whether the input function is growing
     *                  or shrinking.
     * @return          Inverse Function
     */
    public static Function invert(final Function function, final double guess,
                                  final double cutoff, final boolean growing)
    {
        final double PRECISION = 1e-15;
        final double cutoff_abs = Math.abs(cutoff);
        final double sign = (growing ? 1.0 : -1.0);
        return new Function()
        {
            public double of (double y)
            {
                double step = cutoff_abs,
                        test = 0.0,
                        out = guess;
                while (step > PRECISION)
                {
                    double diff = 1.0;
                    while (diff > 0)
                    {
                        out = test;
                        test = out + step;
                        if (test == out)
                        {
                            break;
                        }
                        diff = sign * (y - function.of(test));
                        if (test > cutoff_abs)
                        {
                            return cutoff_abs;
                        }
                    }
                    test = out;
                    step *= 0.5;
                }
                if (out == guess)
                {
                    step = -cutoff_abs;
                    test = 0.0;
                    while (-step > PRECISION)
                    {
                        double diff = 1.0;
                        while (diff > 0)
                        {
                            out = test;
                            test = out + step;
                            if (test == out)
                            {
                                break;
                            }
                            diff = sign * (function.of(test) - y);
                            if (test < -cutoff_abs)
                            {
                                return -cutoff_abs;
                            }
                        }
                        test = out;
                        step *= 0.5;
                    }
                }
                return out;
            }
        };
    }

    /**
     * Implementation of the probit function with double precision following
     * 
     * "Algorithm AS 241: The Percentage Points of the Normal Distribution"
     * Michael J. Wichura
     * Journal of the Royal Statistical Society. Series C (Applied Statistics)
     * Vol. 37, No. 3 (1988), pp. 477-484
     * 
     * The function calculates values of approximating polynomials with
     * pre-calculated coefficients. It is not optimized for performance
     * but for source code readability.
     */
    public static double probit(final double x) throws IllegalArgumentException
    {
        // Coefficients for x close to 0.5
        final double[] a1 = new double[]{3.3871328727963666080e0,
                                         1.3314166789178437745e+2,
                                         1.9715909503065514427e+3,
                                         1.3731693765509461125e+4,
                                         4.5921953931549871457e+4,
                                         6.7265770927008700853e+4,
                                         3.3430575583588128105e+4,
                                         2.5090809287301226727e+3},
                       a2 = new double[]{1.0,
                                         4.2313330701600911252e+1,
                                         6.8718700749205790830e+2,
                                         5.3941960214247511077e+3,
                                         2.1213794301586595867e+4,
                                         3.9307895800092710610e+4,
                                         2.8729085735721942674e+4,
                                         5.2264952788528545610e+3},
        // Coefficients for x in intermediate region.
                       b1 = new double[]{1.42343711074968357734e0,
                                         4.63033784615654529590e0,
                                         5.76949722146069140550e0,
                                         3.64784832476320460504e0,
                                         1.27045825245236838258e0,
                                         2.41780725177450611770e-1,
                                         2.27238449892691845833e-2,
                                         7.74545014278341407640e-4},
                       b2 = new double[]{1.0,
                                         2.05319162663775882187e0,
                                         1.67638483018380384940e0,
                                         6.89767334985100004550e-1,
                                         1.48103976427480074590e-1,
                                         1.51986665636164571966e-2,
                                         5.47593808499534494600e-4,
                                         1.05075007164441684324e-9},
        // Coefficients for x near 0 or 1.
                       c1 = new double[]{6.65790464350110377720e0,
                                         5.46378491116411436990e0,
                                         1.78482653991729133580e0,
                                         2.96560571828504891230e-1,
                                         2.65321895265761230930e-2,
                                         1.24266094738807843860e-3,
                                         2.71155556874348757815e-5,
                                         2.01033439929228813265e-7},
                       c2 = new double[]{1.0,
                                         5.99832206555887937690e-1,
                                         1.36929880922735805310e-1,
                                         1.48753612908506148525e-2,
                                         7.86869131145613259100e-4,
                                         1.84631831751005468180e-5,
                                         1.42151175831644588870e-7,
                                         2.04426310338993978564e-15};
        // Other constants
        final double split1 = 0.425, split2 = 5.0, const1 = 0.180625, const2 = 1.6,
                     y = x - 0.5;

        // Determine region and calculate polynomial values.
        double z, out;
        if (Math.abs(y) <= split1)
        {
            z = const1 - y * y;
            return y * polynomial(a1, z) / polynomial(a2, z);
        }
        else
        {
            if (y < 0)
            {
                z = x;
            }
            else
            {
                z = 1.0 - x;
            }
            if (z <= 0.0)
            {
                throw new IllegalArgumentException("" + x);
            }

            z = Math.sqrt(-Math.log(z));
            if (z <= split2)
            {
                z -= const2;
                out = polynomial(b1, z) / polynomial(b2, z);
            }
            else
            {
                z -= split2;
                out = polynomial(c1, z) / polynomial(c2, z);
            }
            if (y < 0.0)
            {
                return -out;
            }
        }
        return out;
    }

    /**
     * An implementation of the Jacobi method. Returns the
     * matrix of Eigenvalues and the matrix of Eigenvectors.
     */
    public static double[][][] diagonalize (double[][] matrix)
    {
        double[][] diag = copy(matrix);
        double[][] vectors = new double[matrix.length][matrix.length];
        for (int i = 0; i < matrix.length; i++)
        {
            vectors[i][i] = 1;
        }

        while (true)
        {
            // Find entry farthest away from 0.
            double max = 0;
            int a = -1, b = -1;   // row and column number for "maximum".
            for (int i = 0; i < diag.length; i++)
            {
                for (int j = i + 1; j < diag.length; j++)
                {
                    if (Math.abs(diag[i][j]) > max)
                    {
                        max = Math.abs(diag[i][j]);
                        a = i;
                        b = j;
                    }
                }
            }

            if (a < 0)
            {
                break;
            }

            double[][] rotation = rotate2x2(new double[][]{{diag[a][a], diag[a][b]},
                                                           {diag[b][a], diag[b][b]}});
            if (Math.abs(rotation[0][1]) < 1e-15)
            {
                break;
            }
            diag = adjoin(diag, rotation, a, b);
            vectors = multiply(vectors, rotation, a, b, false);
        }
        return new double[][][]{diag, vectors};
    }

    public static boolean[][] dilate(boolean[][] binary_image, boolean[][] mask)
    {
        if (binary_image == null || binary_image.length == 0 || binary_image[0].length == 0 ||
            mask == null || mask.length % 2 == 0 || mask[0].length != mask.length)
        {
            System.err.println("FilamentSensor CellData.dilate: " +
                               "Image and mask must exist. Mask must be quadratic " +
                               "and its side length must be an odd number of pixels.");
            return null;
        }

        final int width = binary_image.length,
                  height = binary_image[0].length,
                  range = mask.length / 2;
        
        boolean[][] out = new boolean[width][height];

        // Dilation
        for (int x = 0; x < width; x++)
        {
            if (x == range)
            {
                x = width - range;
            }
            for (int y = 0; y < height; y++)
            {
                if (binary_image[x][y])
                {
                    int x_min = - Math.min(x, range),
                        x_max = Math.min(width-x-1, range),
                        y_min = - Math.min(y, range),
                        y_max = Math.min(height-y-1, range);

                    for (int a = x_min; a <= x_max; a++)
                    {
                        for (int b = y_min; b <= y_max; b++)
                        {
                            out[x+a][y+b] = mask[a + range][b + range] || out[x+a][y+b];
                        }
                    }
                }
            }
        }
        
        for (int y = 0; y < height; y++)
        {
            if (y == range)
            {
                y = height - range;
            }
            for (int x = 0; x < width; x++)
            {

                if (binary_image[x][y])
                {
                    int x_min = - Math.min(x, range),
                        x_max = Math.min(width-x-1, range),
                        y_min = - Math.min(y, range),
                        y_max = Math.min(height-y-1, range);

                    for (int a = x_min; a <= x_max; a++)
                    {
                        for (int b = y_min; b <= y_max; b++)
                        {
                            out[x+a][y+b] = mask[a + range][b + range] || out[x+a][y+b];
                        }
                    }
                }
            }
        }
        for (int x = range; x < width - range; x++)
        {
            for (int y = range; y < height - range; y++)
            {
                if (binary_image[x][y])
                {
                    if (!binary_image[x-1][y] || !binary_image[x+1][y] ||
                        !binary_image[x][y-1] || !binary_image[x][y+1])
                    {
                        for (int a = -range; a <= range; a++)
                        {
                            for (int b = -range; b <= range; b++)
                            {
                                out[x+a][y+b] = mask[a + range][b + range] || out[x+a][y+b];
                            }
                        }
                    }
                    else
                    {
                        out[x][y] = true;
                    }
                }
            }
        }
        return out;
    }

    public static boolean[][] close(boolean[][] binary_image, boolean[][] mask)
    {
        if (binary_image == null || binary_image.length == 0 || binary_image[0].length == 0 ||
            mask == null || mask.length % 2 == 0 || mask[0].length != mask.length)
        {
            System.err.println("FilamentSensor Calc.close: " +
                               "Image and mask must exist. Mask must be quadratic " +
                               "and its side length must be an odd number of pixels.");
            return null;
        }

        final int width = binary_image.length,
                  height = binary_image[0].length,
                  range = mask.length / 2,
                  width_plus = width + 2 * range,
                  height_plus = height + 2 * range;
        
        boolean[][] tmp = new boolean[width_plus][height_plus];

        // Dilation
        for (int x = 0; x < width; x++)
        {
            if (binary_image[x][0])
            {
                for (int a = 0; a <= 2 * range; a++)
                {
                    for (int b = 0; b <= 2 * range; b++)
                    {
                        tmp[x+a][b] = mask[a][b] || tmp[x+a][b];
                    }
                }
            }
            if (binary_image[x][height-1])
            {
                for (int a = 0; a <= 2 * range; a++)
                {
                    for (int b = 0; b <= 2 * range; b++)
                    {
                        tmp[x+a][height-1+b] = mask[a][b] || tmp[x+a][height-1+b];
                    }
                }
            }
        }
        for (int y = 0; y < height; y++)
        {
            if (binary_image[0][y])
            {
                for (int a = 0; a <= 2 * range; a++)
                {
                    for (int b = 0; b <= 2 * range; b++)
                    {
                        tmp[a][y+b] = mask[a][b] || tmp[a][y+b];
                    }
                }
            }
            if (binary_image[width-1][y])
            {
                for (int a = 0; a <= 2 * range; a++)
                {
                    for (int b = 0; b <= 2 * range; b++)
                    {
                        tmp[width-1+a][y+b] = mask[a][b] || tmp[width-1+a][y+b];
                    }
                }
            }
        }

        for (int x = 1; x < width - 1; x++)
        {
            for (int y = 1; y < height - 1; y++)
            {
                if (binary_image[x][y])
                {
                    if (!binary_image[x-1][y] || !binary_image[x+1][y] ||
                        !binary_image[x][y-1] || !binary_image[x][y+1])
                    {
                        for (int a = 0; a <= 2 * range; a++)
                        {
                            for (int b = 0; b <= 2 * range; b++)
                            {
                                tmp[x+a][y+b] = mask[a][b] || tmp[x+a][y+b];
                            }
                        }
                    }
                    else
                    {
                        tmp[x+range+1][y+range+1] = true;
                    }
                }
            }
        }

        // Copy dilated image to original
        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                binary_image[x][y] = tmp[x + range][y + range];
            }
        }

        // Erosion
        for (int x = 0; x < width_plus; x++)
        {
            if (x == 2 * range)
            {
                x = width;
            }
            for (int y = 0; y < height_plus; y++)
            {
                // Left and right boundary
                if (!tmp[x][y] &&
                    ((x > 0 && tmp[x-1][y]) || (x < width_plus - 1 && tmp[x+1][y]) ||
                     (y > 0 && tmp[x][y-1]) || (y < height_plus - 1 && tmp[x][y+1])))
                {
                    for (int a = 0; a <= 2 * range; a++)
                    {
                        if (x-a >= 0 && x-a < width)
                        {
                            for (int b = 0; b <= 2 * range; b++)
                            {
                                if (y-b >= 0 && y-b < height && mask[a][b])
                                {
                                    binary_image[x-a][y-b] = false;
                                }
                            }
                        }
                    }
                }
            }
        }
        for (int y = 0; y < height_plus; y++)
        {
            if (y == 2 * range)
            {
                y = height;
            }
            for (int x = 0; x < width_plus; x++)
            {
                // Upper and lower boundary
                if (!tmp[x][y] &&
                    ((x > 0 && tmp[x-1][y]) || (x < width_plus - 1 && tmp[x+1][y]) ||
                     (y > 0 && tmp[x][y-1]) || (y < height_plus - 1 && tmp[x][y+1])))
                {
                    for (int b = 0; b <= 2 * range; b++)
                    {
                        if (y-b >= 0 && y-b < height)
                        {
                            for (int a = 0; a <= 2 * range; a++)
                            {
                                if (x-a >= 0 && x-a < width && mask[a][b])
                                {
                                    binary_image[x-a][y-b] = false;
                                }
                            }
                        }
                    }
                }
            }
        }
        // Interior
        for (int x = 2 * range; x < width; x++)
        {
            for (int y = 2 * range; y < height; y++)
            {
                if (!tmp[x][y] &&
                    (tmp[x-1][y] || tmp[x+1][y] || tmp[x][y-1] || tmp[x][y+1]))
                {
                    for (int a = 0; a <= 2 * range; a++)
                    {
                        for (int b = 0; b <= 2 * range; b++)
                        {
                            binary_image[x-a][y-b] = binary_image[x-a][y-b] && !mask[a][b];
                        }
                    }
                }
            }
        }
        return binary_image;
    }

    public static boolean[][] erode(boolean[][] binary_image, boolean[][] mask)
    {
        if (binary_image == null || binary_image.length == 0 || binary_image[0].length == 0 ||
            mask == null || mask.length % 2 == 0 || mask[0].length != mask.length)
        {
            System.err.println("FilamentSensor Calc.erode: " +
                               "Image and mask must exist. Mask must be quadratic " +
                               "and its side length must be an odd number of pixels.");
            return null;
        }

        final int width = binary_image.length,
                  height = binary_image[0].length,
                  range = mask.length / 2;
        
        boolean[][] tmp = new boolean[width][height];
        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                tmp[x][y] = binary_image[x][y];
            }
        }

        for (int x = 0; x < width; x++)
        {
            if (x == 2 * range)
            {
                x = width - range;
            }
            for (int y = 0; y < height; y++)
            {
                // Left and right boundary
                if (!tmp[x][y] &&
                    ((x > 0 && tmp[x-1][y]) || (x < width - 1 && tmp[x+1][y]) ||
                     (y > 0 && tmp[x][y-1]) || (y < height - 1 && tmp[x][y+1])))
                {
                    for (int a = 0; a <= 2 * range; a++)
                    {
                        int x_a = x + a - range;
                        if (x_a >= 0 && x_a < width)
                        {
                            for (int b = 0; b <= 2 * range; b++)
                            {
                                int y_b = y + b - range;
                                if (y_b >= 0 && y_b < height && mask[a][b])
                                {
                                    binary_image[x_a][y_b] = false;
                                }
                            }
                        }
                    }
                }
            }
        }
        for (int y = 0; y < height; y++)
        {
            if (y == range)
            {
                y = height - range;
            }
            for (int x = 0; x < width; x++)
            {
                // Upper and lower boundary
                if (!tmp[x][y] &&
                    ((x > 0 && tmp[x-1][y]) || (x < width - 1 && tmp[x+1][y]) ||
                     (y > 0 && tmp[x][y-1]) || (y < height - 1 && tmp[x][y+1])))
                {
                    for (int b = 0; b <= 2 * range; b++)
                    {
                        int y_b = y + b - range;
                        if (y_b >= 0 && y_b < height)
                        {
                            for (int a = 0; a <= 2 * range; a++)
                            {
                                int x_a = x + a - range;
                                if (x_a >= 0 && x_a < width && mask[a][b])
                                {
                                    binary_image[x_a][y_b] = false;
                                }
                            }
                        }
                    }
                }
            }
        }
        // Interior
        for (int x = range; x < width - range; x++)
        {
            for (int y = range; y < height - range; y++)
            {
                if (!tmp[x][y] &&
                    (tmp[x-1][y] || tmp[x+1][y] || tmp[x][y-1] || tmp[x][y+1]))
                {
                    for (int a = 0; a <= 2 * range; a++)
                    {
                        for (int b = 0; b <= 2 * range; b++)
                        {
                            int x_a = x + a - range, y_b = y + b - range;
                            binary_image[x_a][y_b] = binary_image[x_a][y_b] && !mask[a][b];
                        }
                    }
                }
            }
        }
        return binary_image;
    }

    private static double polynomial(final double[] coefficients,  final double argument)
    {
        final int n = coefficients.length - 1;
        double out = coefficients[n];
        for (int i = n - 1; i >= 0; i--)
        {
            out *= argument;
            out += coefficients[i];
        }
        return out;
    }

    private static double[][] rotate2x2 (double[][] matrix_2x2)
    {
        double tmp = (matrix_2x2[1][1] - matrix_2x2[0][0]) / (2 * matrix_2x2[0][1]);
        double y = tmp + Math.sqrt(1 + tmp * tmp);
        double x = 1 / Math.sqrt(1 + y * y);
        y *= x;
        return new double[][]{{x,-y},{y,x}};
    }

    private static double[][] adjoin (double[][] matrix, double[][] matrix_2x2, int a, int b)
    {
        double[][] ad_matrix_2x2 = new double[][]{{matrix_2x2[0][0], matrix_2x2[1][0]},
                                                  {matrix_2x2[0][1], matrix_2x2[1][1]}};        
        return multiply(multiply(matrix, ad_matrix_2x2, a, b, true), matrix_2x2, a, b, false);
    }

    private static double[][] multiply (double[][] matrix, double[][] matrix_2x2, int a, int b, boolean sparse_left)
    {
        double[][] out = copy(matrix);
        for (int i = 0; i < matrix.length; i++)
        {
            if (sparse_left)
            {
                out[a][i] = matrix_2x2[0][0] * matrix[a][i] + matrix_2x2[0][1] * matrix[b][i];
                out[b][i] = matrix_2x2[1][0] * matrix[a][i] + matrix_2x2[1][1] * matrix[b][i];
            }
            else
            {
                out[i][a] = matrix[i][a] * matrix_2x2[0][0] + matrix[i][b] * matrix_2x2[1][0];
                out[i][b] = matrix[i][a] * matrix_2x2[0][1] + matrix[i][b] * matrix_2x2[1][1];
            }
        }
        return out;
    }

    private static double[][] copy (double[][] matrix)
    {
        double[][] copy = new double[matrix.length][matrix.length];
        for (int i = 0; i < matrix.length; i++)
        {
            for (int j = 0; j < matrix.length; j++)
            {
                copy[i][j] = matrix[i][j];
            }
        }
        return copy;
    }

}
