package core;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.awt.Point;
import java.util.List;

import org.w3c.dom.Element;

import ui.FilamentSensor;

public class CellData implements Cloneable
{

    // Orientation, aspect ratio, and order parameter
    // are scaled by Const.M.
    private long m_orientation, m_aspect_ratio, m_order_parameter;
    private int m_area;
    private Point m_center;
    private boolean[][] m_binary_image;

    CellData(boolean[][] binary_image)
    {
        binary_image = Calc.close(binary_image, Calc.circleMask(20));
        binary_image = remove_spots(binary_image, 15);
        final int width = binary_image.length,
                  height = binary_image[0].length;
        double sum_x = 0, sum_y = 0, sum_xx = 0, sum_xy = 0, sum_yy = 0;
        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                if (binary_image[x][y])
                {
                    m_area++;
                    sum_x += x;
                    sum_y += y;
                    sum_xx += x * x;
                    sum_xy += x * y;
                    sum_yy += y * y;
                }
            }
        }
        double area = (double) m_area;
        int mean_x = (int)Math.round(sum_x / area);
        int mean_y = (int)Math.round(sum_y / area);
        double cov_xx = sum_xx / area - sum_x * sum_x / (area * area);
        double cov_xy = sum_xy / area - sum_x * sum_y / (area * area);
        double cov_yy = sum_yy / area - sum_y * sum_y / (area * area);
        double ew_hi = (cov_xx + cov_yy) / 2.0 +
                       Math.sqrt((cov_xx - cov_yy) * (cov_xx - cov_yy) / 4.0 + cov_xy * cov_xy);
        double ew_lo = (cov_xx + cov_yy) / 2.0 -
                       Math.sqrt((cov_xx - cov_yy) * (cov_xx - cov_yy) / 4.0 + cov_xy * cov_xy);
        double phi = 180 - Math.atan2(ew_hi - cov_xx, cov_xy) / Const.RAD;

        if (FilamentSensor.DEBUG)
        {
            System.out.println("Cell center: (" + mean_x + ", " + mean_y +")");
            System.out.format("Covariance matrix: %n%.3f %.3f%n%.3f %.3f%n", cov_xx, cov_xy, cov_xy, cov_yy);
            System.out.format("Eigenvalues: %.3f %.3f%nAspect ratio: %.3f%nOrientation: %.3f%n",
                               ew_hi, ew_lo, Math.sqrt(ew_hi / ew_lo), phi);
        }

        m_orientation = (long) (Const.M * phi);
        m_aspect_ratio = (long) (Const.M * Math.sqrt(ew_hi / ew_lo));
        m_center = new Point(mean_x, mean_y);
        m_order_parameter = -1000 * Const.M;
        m_binary_image = Calc.dilate(binary_image, Calc.circleMask(10));
    }

    private CellData(long phi, long aspect_ratio, long order_parameter,
                     int area, Point center, boolean[][] binary_image)
    {
        m_orientation = phi;
        m_aspect_ratio = aspect_ratio;
        m_order_parameter = order_parameter;
        m_area = area;
        m_center = (Point)center.clone();
        if (binary_image != null)
        {
            final int width = binary_image.length,
                      height = binary_image[0].length;
            m_binary_image = new boolean[width][height];
            for (int x = 0; x < width; x++)
            {
                for (int y = 0; y < height; y++)
                {
                    m_binary_image[x][y] = binary_image[x][y];
                }
            }
        }
    }

    public CellData(Element element)
    {
        this(Long.parseLong(element.getAttribute("orientation")),
             Long.parseLong(element.getAttribute("aspect_ratio")),
             Long.parseLong(element.getAttribute("order_parameter")),
             Integer.parseInt(element.getAttribute("area")),
             new Point(Integer.parseInt(element.getAttribute("center_x")),
                       Integer.parseInt(element.getAttribute("center_y"))), null);
    }

    void calculateOrderParameter (List<FilamentChain> list)
    {
        double numerator = 0.0,
               denominator = 0.0;

        for (FilamentChain chain : list)
        {
            double pixels = chain.getLength() * chain.getWidth(),
                   phi = (double) (chain.getOrientation() - m_orientation) / Const.MF;
            numerator += pixels * Math.cos(2.0 * phi * Const.RAD);
            denominator += pixels;
        }
        m_order_parameter = Math.round(Const.MF * numerator / denominator);
    }

    void shutdown()
    {
        m_binary_image = null;
    }

    public Point getCenter()
    {
        return m_center;
    }

    public long getOrientation()
    {
        return m_orientation;
    }

    public long getAspectRatio()
    {
        return m_aspect_ratio;
    }

    public long getOrderParameter()
    {
        return m_order_parameter;
    }

    public long getLongHalfAxis()
    {
        return (long)Math.sqrt(m_area * Const.MF * m_aspect_ratio / Math.PI);
    }

    public int getArea()
    {
        return m_area;
    }

    boolean[][] getMask()
    {
        if (m_binary_image == null)
        {
            return null;
        }
        final int width = m_binary_image.length,
                  height = m_binary_image[0].length;
        boolean[][] mask = new boolean[width][height];
        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                mask[x][y] = m_binary_image[x][y];
            }
        }
        return mask;
    }

    private static boolean[][] remove_spots(boolean[][] binary_image, int size)
    {
        if (binary_image == null || binary_image.length == 0 || binary_image[0].length == 0)
            {
                System.err.println("FilamentSensor CellData.remove_spots: " +
                                   "Image must exist and have non-zero size.");
                return null;
            }

        final int width = binary_image.length,
                  height = binary_image[0].length;

        // Upper and lower boundary
        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                if (y == size)
                {
                    y = height - size;
                }
                if (binary_image[x][y])
                {
                    if ((x >= size          && binary_image[x-size][y]) ||
                        (x < width - size  && binary_image[x+size][y]) ||
                        (y >= size          && binary_image[x][y-size]) ||
                        (y < height - size && binary_image[x][y+size]))
                    {
                        continue;
                    }
                    
                    boolean bright_pixel = false;
                    for (int i = 1; i <= size; i++)
                    {
                        int[][] points = new int[][]{
                            {x-size, y-i}, {x-size, y+i}, {x-i, y-size}, {x-i, y+size},
                            {x+i, y-size}, {x-i, y+size}, {x+size, y-i}, {x+size, y+i}};
                        for (int[] p : points)
                        {
                            if (p[0] > 0 && p[0] < width && p[1] > 0 && p[1] < height)
                            {
                                bright_pixel = bright_pixel || binary_image[p[0]][p[1]];
                            }
                        }
                        if (bright_pixel)
                        {
                            break;
                        }
                    }

                    if (!bright_pixel)
                    {
                        int x_min = Math.max(0, x-size), x_max = Math.min(width, x+size+1),
                            y_min = Math.max(0, y-size), y_max = Math.min(height, y+size+1);
                        for (int a = x_min; a < x_max; a++)
                        {
                            for (int b = y_min; b < y_max; b++)
                            {
                                binary_image[a][b] = false;
                            }
                        }
                    }
                }
            }
        }

        // Left and right boundary
        for (int x = 0; x < width; x++)
        {
            if (x == size)
            {
                x = width - size;
            }
            for (int y = 0; y < height; y++)
            {
                if (binary_image[x][y])
                {
                    if ((x >= size         && binary_image[x-size][y]) ||
                        (x < width - size  && binary_image[x+size][y]) ||
                        (y >= size         && binary_image[x][y-size]) ||
                        (y < height - size && binary_image[x][y+size]))
                    {
                        continue;
                    }
                    
                    boolean bright_pixel = false;
                    for (int i = 1; i <= size; i++)
                    {
                        int[][] points = new int[][]{
                            {x-size, y-i}, {x-size, y+i}, {x-i, y-size}, {x-i, y+size},
                            {x+i, y-size}, {x-i, y+size}, {x+size, y-i}, {x+size, y+i}};
                        for (int[] p : points)
                        {
                            if (p[0] > 0 && p[0] < width && p[1] > 0 && p[1] < height)
                            {
                                bright_pixel = bright_pixel || binary_image[p[0]][p[1]];
                            }
                        }
                        if (bright_pixel)
                        {
                            break;
                        }
                    }

                    if (!bright_pixel)
                    {
                        int x_min = Math.max(0, x-size), x_max = Math.min(width, x+size+1),
                            y_min = Math.max(0, y-size), y_max = Math.min(height, y+size+1);
                        for (int a = x_min; a < x_max; a++)
                        {
                            for (int b = y_min; b < y_max; b++)
                            {
                                binary_image[a][b] = false;
                            }
                        }
                    }
                }
            }
        }

        // Interior
        for (int x = size; x < width - size; x++)
        {
            for (int y = size; y < height - size; y++)
            {
                if (binary_image[x][y])
                {
                    if (binary_image[x-size][y] || binary_image[x+size][y] ||
                        binary_image[x][y-size] || binary_image[x][y+size])
                    {
                        continue;
                    }
                    
                    boolean bright_pixel = false;
                    for (int i = 1; i <= size; i++)
                    {
                        int[][] points = new int[][]{
                            {x-size, y-i}, {x-size, y+i}, {x-i, y-size}, {x-i, y+size},
                            {x+i, y-size}, {x-i, y+size}, {x+size, y-i}, {x+size, y+i}};
                        for (int[] p : points)
                        {
                            bright_pixel = bright_pixel || binary_image[p[0]][p[1]];
                        }
                        if (bright_pixel)
                        {
                            break;
                        }
                    }

                    if (!bright_pixel)
                    {
                        for (int a = x-size; a < x+size; a++)
                        {
                            for (int b = y-size; b < y+size; b++)
                            {
                                binary_image[a][b] = false;
                            }
                        }
                    }
                }
            }
        }
        return binary_image;
    }

    @Override
    public Object clone()
    {
        return new CellData(m_orientation, m_aspect_ratio, m_order_parameter,
                             m_area, m_center, m_binary_image);
    }

}
