package core;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormatSymbols;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.Set;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.image.BufferedImage;

import javax.imageio.ImageIO;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.event.ChangeListener;

import core.DataParameters.Any;
import ui.Config;
import ui.Dialogs;
import ui.FilamentSensor;
import ui.ImageFactory;
import ui.MainWindow;

public class DataImage extends DataProject implements Misc
{
    private File m_image_file;

    private ImageProcessing m_image_processing;
    private DataParameters m_parameter_container;
    private DataFilaments m_filament_container;

    private LineSensor m_line_sensor;

    private String m_image_name, m_image_resolution;
    private static final String ls = System.lineSeparator();

    private CellData m_cell_data;

    public static final int ORIG = 0, PRE = 1, ELOG_COLOR = 2, ELOG_GREY = 3, DIFF = 4, BIN = 5,
                            ORI_MAP = 6,  HOUGH = 7;

    // Main constructor. For reconstruction from save files.
    public DataImage(File img_file, CellData cell_data, DataParameters parameters,
                     DataFilaments filaments)
    {
        m_image_file = img_file;

        if (img_file == null)
        {
            m_image_processing = null;
            m_cell_data = cell_data;
            m_image_name = "";
        }
        else
        {
            m_image_processing = new ImageProcessing(img_file);
            m_cell_data = (cell_data == null ? m_image_processing.getCellData() : cell_data);
            m_image_name = img_file.getName();
            if (FilamentSensor.DEBUG)
            {
                System.out.println("Processing with ImageProcessing was successful.");
            }
        }

        m_parameter_container = (parameters == null ? getDefaultParameters() : parameters);

        m_filament_container = (filaments == null ? new DataFilaments() : filaments);
        
        if (filaments != null && m_cell_data != null)
        {
            m_cell_data.calculateOrderParameter(m_filament_container.getFilamentChains());
        }

        m_line_sensor = new LineSensor();

        addAsProcessed();
    }

    // Constructor for new files.
    public DataImage(File img_file, DataParameters parameters)
    {
        this(img_file, null, parameters, null);
    }

    /**
     * Sets the ImageProcessing and LineSensor members to null in order to save memory.
     */
    protected void shutDown()
    {
        m_image_processing = null;
        m_line_sensor = null;
        m_cell_data.shutdown();
    }

    /**
     * This method adds this DataImage object to the list of processed images,
     * that will be saved when saveProject is invoked. If a DataImage object for
     * the same image file already appears in the list of processed images, that
     * older object is discarded from the list, to avoid saving data for one
     * image twice.
     *
     * To uniquely identify the image to which a DataImage object belongs, the
     * content of the name field of the DataImage class is used.
     */
    void addAsProcessed()
    {
        for (int i = m_processed_images.size() - 1; i >= 0; i--)
        {
            DataImage tracked_change = m_processed_images.get(i);

            if (m_image_name.equals(tracked_change.getImageName()))
            {
                m_processed_images.remove(tracked_change);
            }
        }
        m_processed_images.add(this);
    }

    public DataParameters parameters()
    {
        if (m_parameter_container == null || m_force_defaults)
        {
            m_parameter_container = getDefaultParameters();
        }
        return m_parameter_container.completeFrom(getDefaultParameters());
    }

    private ImageProcessing imageProcessing()
    {
        if (m_image_processing == null)
        {
            m_image_processing = new ImageProcessing(m_image_file);
        }
        return m_image_processing;
    }

    private LineSensor lineSensor()
    {
        if (m_line_sensor == null)
        {
            m_line_sensor = new LineSensor();
        }
        return m_line_sensor;
    }

    private CellData cellData()
    {
        if (m_cell_data == null)
        {
            m_cell_data = imageProcessing().getCellData();
        }
        return m_cell_data;
    }

    private boolean[][] areaMask()
    {
        if (m_cell_data == null || m_cell_data.getMask() == null)
        {
            m_cell_data = imageProcessing().getCellData();
        }
        return m_cell_data.getMask();
    }

    /**
     * Get the name of the image file corresponding to this data file.
     *
     * @return A string containing the name of the image file.
     */
    public String getImageName()
    {
        return m_image_name;
    }

    /**
     * Get the image file corresponding to this data file.
     *
     * @return The image file.
     */
    public File getFile()
    {
        return m_image_file;
    }

    public String getResolution()
    {
        if (m_image_processing != null)
        {
            m_image_resolution = m_image_processing.getResolution();
        }
        return m_image_resolution;
    }

    public CellData getCellData()
    {
        return (CellData) cellData().clone();
    }

    public boolean usesDefaultParameters()
    {
        return parameters().equals(m_default_parameters);
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////    Export functions for text files     /////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * The import from CSV assumes orientations counterclockwise from
     * version 0.1.3 on. This implies "inverting the y-axis". When older
     * CSV files are imported, filament end points will be wrong. The
     * filament class has a method to remedy this.
     */
    public boolean importCSV()
    {
        File csv_file = Dialogs.load(false, "Choose csv file.");
        List<String> input = new ArrayList<String>();
        try
        {
            Scanner scanner = new Scanner(csv_file);
            while (scanner.hasNext("#center_x,center_y,length,angle,width"))
            {
                if (FilamentSensor.DEBUG)
                {
                    System.out.println("Skipping line.");
                }
                scanner.nextLine();
            }
            while (scanner.hasNext())
            {
                if (FilamentSensor.DEBUG)
                {
                    System.out.println("Reading line.");
                }
                input.add(scanner.nextLine());
            }
            for (String line : input)
            {
                String dec = Character.toString((new DecimalFormatSymbols()).getDecimalSeparator());
                line = line.replaceAll(",", "_").replaceAll("\\.", dec);
                scanner = new Scanner(line);
                scanner.useDelimiter("_");
                if (scanner.hasNextDouble())
                {
                    double center_x = scanner.nextDouble();
                    double center_y = scanner.nextDouble();
                    double length = scanner.nextDouble();
                    double angle = scanner.nextDouble();
                    long width = Math.round(Const.MF * scanner.nextDouble());
                    int head_x = (int)Math.round(center_x + length * Math.cos(angle * Const.RAD) / 2.0);
                    int head_y = (int)Math.round(center_y - length * Math.sin(angle * Const.RAD) / 2.0);
                    int tail_x = (int)Math.round(center_x - length * Math.cos(angle * Const.RAD) / 2.0);
                    int tail_y = (int)Math.round(center_y + length * Math.sin(angle * Const.RAD) / 2.0);
                    Filament filament = new Filament(new Point(head_x, head_y),
                                                     new Point(tail_x, tail_y),
                                                     width, true);
                    m_filament_container.addFilamentChain(new FilamentChain(filament), false);
                }
                
            }
            scanner.close();
        }
        catch(FileNotFoundException e)
        {
            System.out.println("File not found." + e.getMessage());
        }

        return false;
    }

    /**
     * Exports the list of filament chains as comma separated values.
     */
    public void exportCSV(File path, boolean chains, boolean dialog)
    {
        String out = "#cell_properties" + 
                     ls + "%center_x" + ls + cellData().getCenter().x +
                     ls + "%center_y" + ls + cellData().getCenter().y +
                     ls + "%area" + ls + cellData().getArea() +
                     ls + "%aspect_ratio" + ls + cellData().getAspectRatio() / Const.MF +
                     ls + "%long_half_axis" + ls + cellData().getLongHalfAxis() / Const.MF +
                     ls + "%orientation" + ls + cellData().getOrientation() / Const.MF +
                     ls + "%order_parameter" + ls + cellData().getOrderParameter() / Const.MF +
                     ls;
        if (m_filament_container.countChains() > 0)
        {
            out += "%filaments" + ls + "#center_x,center_y,length,angle,width" + ls;
            for (FilamentChain fil : m_filament_container.getFilamentChains())
            {
                out += fil.getCenter().x / 10.0 + "," + fil.getCenter().y / 10.0 + "," + fil.getLength() / Const.MF + ","
                        + fil.getOrientation() / Const.MF + "," + fil.getWidth() / Const.MF + ls;
            }
        }

        export("Filament data to export", out, dialog, path, chains);
    }

    public void exportHistogram(File path, boolean dialog, String parameters, int type)
    {
        String type_name = "";
        switch (type)
        {
        case ELOG_COLOR:
            type_name = "eLoG";
            break;
        case HOUGH:
            type_name = "Hough transformation";
            break;
        default:
            return;
        }

        if (imageProcessing().getImage(type) == null)
        {
            if (FilamentSensor.DEBUG)
            {
                System.out.println(type_name + "s have not been calculated yet...");
            }
            return;
        }

        String out = parameters + ls + "%orientation_histogram" + ls + "#orientation,pixels" + ls;
        out += imageProcessing().getHistogram(type);

        export(type_name + " histogram", out, dialog, path, false);
    }

    private void export(String description, String content, boolean dialog, File path, boolean chains)
    {
        if (dialog)
        {
            Object option[] = new Object[]
            { "OK", "Save as..." };
            JScrollPane message = new JScrollPane(new JTextArea(content));
            message.setMaximumSize(new Dimension(400, 300));
            message.setPreferredSize(new Dimension(400, 300));

            // Continue only if "Save as..." is clicked in the dialog.
            if (JOptionPane.showOptionDialog(MainWindow.frame(), message, description, JOptionPane.YES_NO_OPTION,
                    JOptionPane.INFORMATION_MESSAGE, null, option, option[0]) != JOptionPane.NO_OPTION)
            {
                return;
            }

            JFileChooser jfc = new JFileChooser();

            String folder = Config.get("LastFolders", "Export");

            if (folder != "")
            {
                jfc.setCurrentDirectory(new File(folder));
            }

            // Continue only if the approving option is clicked in the dialog.
            if (jfc.showSaveDialog(MainWindow.frame()) != JFileChooser.APPROVE_OPTION)
            {
                return;
            }

            if (!writeFile(jfc.getSelectedFile(), content))
            {
                return;
            }

            try
            {
                folder = jfc.getSelectedFile().getParentFile().getCanonicalPath();
                Config.set("LastFolders", "Export", folder);
            }
            catch (IOException e)
            {
            }
        }
        else
        {
            File file = new File(path.getAbsolutePath() + File.separator +
                                 m_image_file.getName().replaceAll("\\.", "_") +
                                 (chains ? "_chains.csv" : "_filaments.csv"));
            writeFile(file, content);
        }
    }

    private boolean writeFile(File file, String content)
    {
        try
        {
            // Default file name for batch processing.
            if (file == null)
            {
                // "[A-Za-z0-9]{3,4}$" should allow for 3 or 4 character file
                // name extensions.
                file = new File(m_image_file.getCanonicalPath().replaceAll("\\.[A-Za-z0-9]{3,4}$", "") + ".csv");
            }

            BufferedWriter bwr = new BufferedWriter(new FileWriter(file, false));
            bwr.write(content);
            bwr.close();

            if (FilamentSensor.DEBUG)
            {
                System.out.println("Exported data to '" + file + "'.");
            }
        }
        catch (IOException e)
        {
            System.out.println("Exporting to file failed: " + e);
            return false;
        }
        return true;
    }

    public boolean saveImage(File path, String identifier, int type, boolean filaments, Color color)
    {
        return saveImage(path, identifier, makeImage(type, filaments, color));
    }

    public boolean saveImage(File path, String identifier, BufferedImage image)
    {
        if (image == null)
        {
            return false;
        }
        File img_file = null;
        if (path == null)
        {
            img_file = Dialogs.save(false, "Save image...");
        }
        else
        {
            img_file = new File(path.getAbsolutePath() + File.separator +
                                m_image_file.getName().replaceAll("\\.", "_") +
                                "_" + identifier + ".png");
            Config.set("LastFolders", "Save", path.getAbsolutePath());
        }

        if (img_file == null)
        {
            return false;
        }

        try
        {
            if (FilamentSensor.DEBUG)
            {
                System.out.println("Try to save image to '" + img_file.getAbsolutePath() + "'.");
            }
            return ImageIO.write(image, "png", img_file);
        }
        catch (IOException e)
        {
            System.out.println(e.getMessage());
            return false;
        }
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////  Exposing the public methods of m_image_processing  ///////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////

    public void updatePreprocessing()
    {
        Map<Any, Integer> pre_params = parameters().getParameters(DataParameters.PRE);
        boolean changed = hasChanged();

        changed = imageProcessing().updatePreprocessing(pre_params, areaMask(), changed);
        setChanged(changed);
    }

    public String calculateELoGs()
    {
        Map<Any, Integer> elog_params = parameters().getParameters(DataParameters.ELOG);
        String out = "#Parameters:" + ls;
        for (Any key : elog_params.keySet())
        {
            out += "%" + key + ls + (key.name() == "orientations" || key.name() == "threshold" ?
                                (int)elog_params.get(key) : elog_params.get(key) / 10.0) + ls;
        }

        if (imageProcessing().calculateELoGs(elog_params))
        {
            setChanged(true);
        }
        return out;
    }

    public String calculateHough()
    {
        Map<Any, Integer> hough_params = parameters().getParameters(DataParameters.HOUGH);
        String out = "#Parameters:" + ls;
        for (Any key : hough_params.keySet())
        {
            out += "%" + key + ls + hough_params.get(key) + ls;
        }

        updateBinarization();
        if (imageProcessing().calculateHough(hough_params))
        {
            setChanged(true);
        }
        return out;
    }

    public void updateBinarization()
    {
        Map<Any, Integer> bin_params = parameters().getParameters(DataParameters.BIN);
        boolean elog_null = true;

        if (imageProcessing().updateBinarization(bin_params, elog_null))
        {
            setChanged(true);
        }
    }

    public int scoreImage()
    {
        return imageProcessing().scoreImage(areaMask(), 3);
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////
    //////////////////  Exposing the public methods of m_parameter_container  //////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////

    public int getParameter(DataParameters.Any key)
    {
        return parameters().getParameter(key);
    }
    
    public int getBlackWhite(boolean get_white)
    {
        if (get_white)
        {
            int white = parameters().getParameter(DataParameters.Pre.white);
            if (white < 0)
            {
                return imageProcessing().getBlackWhite()[1];
            }
            return white;
        }
        int black = parameters().getParameter(DataParameters.Pre.black);
        if (black < 0)
        {
            return imageProcessing().getBlackWhite()[0];
        }
        return black;
    }

    public int getElogThreshold()
    {
        int threshold = parameters().getParameter(DataParameters.Elog.threshold);
        if (threshold < 0)
        {
            return imageProcessing().getElogThreshold();
        }
        return threshold;
    }

    public void setParameter(DataParameters.Any key, int number)
    {
        parameters().setParameter(key, number);
    }
    
    public Set<Any> keySet(int type)
    {
        return parameters().keySet(type);
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////  Exposing the public methods of m_filament_container  //////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////

    public void updateTraceParams()
    {
        Map<Any, Integer> trace_params = parameters().getParameters(DataParameters.TRACE);

        m_filament_container.updateTraceParams(trace_params);
    }

    public void cleanTracedFilaments()
    {
        m_filament_container.cleanTracedFilaments(true);
    }

    public void keepDiscardAllFilaments()
    {
        m_filament_container.keepDiscardAllFilaments();
    }

    public void remedyOldAngles()
    {
        m_filament_container.remedyOldAngles();
    }

    public Pair<Int2D, Integer> calculateWidthMap(double tolerance)
    {
        LineSensor sensor = lineSensor();
        int max = sensor.calcWidthMap(m_image_processing.getBinaryImage(), tolerance);
        return new Pair<>(new Int2D(sensor.getWidthMap()), max);
    }

    public void scanFilaments()
    {
        LineSensor sensor = lineSensor();
        boolean[][] binary_image = imageProcessing().getBinaryImage();
        Map<Any, Integer> trace_params = parameters().getParameters(DataParameters.TRACE);

        m_filament_container.scanFilaments(sensor, binary_image, trace_params);
        cellData().calculateOrderParameter(m_filament_container.getFilamentChains());
        addAsProcessed();
    }

    public void setChanged(boolean changed)
    {
        m_filament_container.setChanged(changed);
    }

    public boolean hasChanged()
    {
        return m_filament_container.hasChanged();
    }

    public void addChangeListener(ChangeListener cl)
    {
        m_filament_container.addChangeListener(cl);
    }

    public List<Filament> getSelectedFilaments()
    {
        return m_filament_container.getSelectedFilaments();
    }

    public List<AbstractFilament> getSelection()
    {
        return m_filament_container.getSelection();
    }

    public void clearSelection()
    {
        m_filament_container.clearSelection();
    }

    public void deleteSelected()
    {
        m_filament_container.deleteSelected();
    }

    public void setSelection(AbstractFilament selected)
    {
        m_filament_container.setSelection(selected);
    }

    public void setSelection(List<? extends AbstractFilament> list)
    {
        m_filament_container.setSelection(list);
    }

    public int countChains()
    {
        return m_filament_container.countChains();
    }

    public int countFilaments()
    {
        return m_filament_container.countFilaments();
    }

    public int countRows()
    {
        return m_filament_container.countRows();
    }

    public AbstractFilament getEntryInRow(int row)
    {
        return m_filament_container.getEntryInRow(row);
    }

    public AbstractFilament changeEntryInRow(int row)
    {
        return m_filament_container.changeEntryInRow(row);
    }

    public int getRowOfEntry(AbstractFilament fil)
    {
        return m_filament_container.getRowOfEntry(fil);
    }
    
    public AbstractFilament wrap(AbstractFilament fil)
    {
        return m_filament_container.wrap(fil);
    }

    public List<Filament> getFilaments()
    {
        return m_filament_container.getFilaments();
    }

    public FilamentChain getFilamentChain(int index)
    {
        return m_filament_container.getFilamentChain(index);
    }

    public List<FilamentChain> getFilamentChains()
    {
        return m_filament_container.getFilamentChains();
    }

    public void addFilamentChain(FilamentChain chain, boolean undoable)
    {
        m_filament_container.addFilamentChain(chain, undoable);
    }
    
    public boolean modifyAbstractFilament(AbstractFilament before, AbstractFilament after)
    {
        return m_filament_container.modifyAbstractFilament(before, after);
    }

    public void undoChain()
    {
        m_filament_container.undoChain();
    }

    public void redoChain()
    {
        m_filament_container.redoChain();
    }

    public BufferedImage makeImage(int type, boolean filaments, Color color)
    {
        BufferedImage image;

        switch (type)
        {
        case ORIG:
        case PRE:
        case ELOG_COLOR:
        case ELOG_GREY:
        case BIN:
        case HOUGH:
            image = imageProcessing().getImage(type);
            break;
        case ORI_MAP:
            image = ImageFactory.makeColorImage(lineSensor().getOrientationField());
            break;
        default:
            return null;
        }

        if (image == null)
        {
            return null;
        }

        if (filaments && (type == ORIG || type == PRE || type == BIN))
        {
            BufferedImage output = new BufferedImage(image.getWidth(), image.getHeight(),
                                                     BufferedImage.TYPE_4BYTE_ABGR);
            Graphics2D graphics2d = output.createGraphics();
            graphics2d.drawImage(image, null, 0, 0);
            graphics2d.setColor(color);

            // draw filaments. TODO: Refactor to use Chains?
            for (Filament filament : m_filament_container.getFilaments())
            {
                graphics2d.setStroke(new BasicStroke((float) (filament.getWidth() / Const.MF),
                                                     BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));
                graphics2d.drawLine(filament.getHead().x, filament.getHead().y,
                                    filament.getTail().x, filament.getTail().y);
            }

            graphics2d.dispose();
            return output;
        }
        return image;
    }

    public BufferedImage makeImage(int type)
    {
        switch (type)
        {
        case ORIG:
        case PRE:
        case ELOG_COLOR:
        case ELOG_GREY:
        case DIFF:
        case BIN:
        case HOUGH:
            return imageProcessing().getImage(type);
        }
        return null;
    }

}
