# -*- coding: utf-8 -*-
"""
Copyright (C) 2014 Benjamin Eltzner

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
or see <http://www.gnu.org/licenses/>.
"""

import numpy as num
import numpy.random as arandom
import random, math, sys
import numbers as n
import operator as op
from scipy.signal import convolve
from PIL import Image as img
from functools import reduce

WIDTH = 800
HEIGHT = 800
PI = num.pi
RAD = math.radians(1)
OUTPUT_FOLDER = 'sim_img'

# This class describes points that can be shifted, copied, and can return the
# nearest lattice point.
class Point(object):
    def __init__(self, x = None, y = None):
        self.x = x
        self.y = y

    def copy(self):
        return Point(self.x, self.y)

    def shift(self, dx, dy):
        self.x += dx
        self.y += dy
        return self

    def asInt(self):
        return Point(int(round(self.x)), int(round(self.y)))

    def coordinates(self):
        return self.x, self.y

    def out(self):
        return "%.1f,%.1f" % (self.x, self.y)


CENTER = Point(WIDTH/2,HEIGHT/2)

# This class describes ellipses with arbitrary orientation in the plane.
class Ellipse(object):
    def __init__(self, center = None, axis1 = None, axis2 = None,
                 orientation = None):
        if isinstance(axis1, n.Real) and isinstance(axis2, n.Real):
            self.shortAxis = min(axis1, axis2)
            self.longAxis = max(axis1, axis2)
        if isinstance(orientation, n.Real):
            self.orientation = orientation % 180
            if self.orientation < 0:
                self.orientation += 180
        self.center = center

    @classmethod
    def random(self, rngAxis1, rngAxis2, rngOrientation):
        area = axis1 = axis2 = 0
        while area < 15000 or max(axis1, axis2) < 180 or area > 35000:
            axis1, axis2 = rngAxis1(), rngAxis2()
            area = axis1 * axis2
        try:
            return Ellipse(CENTER, axis1, axis2, rngOrientation())
        except:
            print("Could not create random Ellipse.")

    # This method returns True, if the point argument is covered by the ellipse.
    def includes(self, point):
        try:
            x = point.x - self.center.x
            y = point.y - self.center.y
            cosPhi = math.cos(self.orientation * RAD)
            sinPhi = math.sin(self.orientation * RAD)
            if ((self.shortAxis * ( x * cosPhi + y * sinPhi))**2 +
                (self.longAxis  * (-x * sinPhi + y * cosPhi))**2 <
                (self.shortAxis * self.longAxis)**2):
                    return True
        except:
            print("Cannot check ellipse inclusion.")
        return False

    def area(self):
        return PI * self.longAxis * self.shortAxis


# This class describes filaments as lines defined by their center point, length
# width and orientation.
class Filament(object):
    def __init__(self, center = None, length = None, width = None,
                 orientation = None):
        self.center = center
        self.length = length
        self.width = width
        self.orientation = ((orientation % 180) + 180) % 180
        try:
            self.neighborhood = Ellipse(self.center, self.length / 2.0 + 10,
                                        self.width / 2.0 + 30, self.orientation)
        except:
            print("Cannot define neighborhood.")
            self.neighborhood = None

    @classmethod
    def random(self, rngCenter, rngLength, rngWidth, rngOrientation):
        try:
            return Filament(rngCenter(), rngLength(), rngWidth(), rngOrientation())
        except:
            print("Could not create random Filament.")

    @classmethod
    def twin(self, filament):
        if random.randint(0,1):
            # rotated
            ori = filament.orientation + makeConstrainedGaussianRNG(5., 2., 3., 10.)()
            center = filament.center
        else:
            # shifted
            ori = filament.orientation
            center = filament.center.copy()
            shift_dist = filament.width * makeConstrainedGaussianRNG(2., 1., 1., 5.)()
            phi = ((ori + 90.) % 180.) * RAD
            center.shift(shift_dist * math.cos(phi), shift_dist * math.sin(phi))
        try:
            return Filament(center, filament.length, filament.width, ori)
        except:
            print("Could not create twin.")

    # This method returns a list of all pixels covered by the filament.
    def pixels(self):
        try:
            if (self.orientation > 45 and self.orientation < 135):
                slope = 1 / (math.tan(self.orientation * RAD))
                center = self.center.copy()
                pixelList = [center.asInt()]
                dx, dy = slope, 1
                point1 = center.copy()
                point2 = center.copy()
                while (dx**2 + dy**2 <= self.length**2 / 4.0):
                    point1.shift(slope, 1)
                    point2.shift(-slope, -1)
                    pixelList.append(point1.asInt())
                    pixelList.append(point2.asInt())
                    dy += 1
                    dx += slope
            else:
                slope = math.tan(self.orientation * RAD)
                center = self.center.copy()
                pixelList = [center.asInt()]
                dx, dy = 1, slope
                point1 = center.copy()
                point2 = center.copy()
                while (dx**2 + dy**2 <= self.length**2 / 4.0):
                    point1.shift(1, slope)
                    point2.shift(-1, -slope)
                    pixelList.append(point1.asInt())
                    pixelList.append(point2.asInt())
                    dx += 1
                    dy += slope
            return pixelList
        except:
            print("Could not get filament pixels.")
        return None

    # This method returns True if the point argument is covered by an elliptic
    # neighborhood of the filament.
    def neighborhoodIncludes(self, point):
        return self.neighborhood.includes(point);

    def out(self):
        return "%s,%.6f,%.6f,%.1f" % (self.center.out(),
                                      self.length + self.width - 1,
                                      180. - self.orientation,
                                      min(max(self.width - 1, 1.), 5.))


# The Cell class contains the method for the assembly of a cell image from
# an ellipse, a list of filaments and several random generators.
class Cell(object):
    def __init__(self, ellipse=None, filaments=None):
        self.ellipse = ellipse
        self.filaments = filaments

    def background(self, ratio):
        pixels = reduce(op.add, [f.pixels() for f in self.filaments])
        pixels = random.sample(pixels, int(ratio * len(pixels)))
        self.image = num.zeros((WIDTH, HEIGHT), num.uint8)
        for p in pixels:
            self.image[p.coordinates()] = 255
        sigma = 1.5
        reach = int(math.ceil(2 * sigma))
        gaussMask = makeGaussMask(sigma, reach)
        for i in range(25):
            self.image = convolve(self.image, gaussMask, mode='same')
            self.image = 255 / num.amax(self.image) * self.image
            self.image[self.image<50] *= 1.1
            self.image[self.image>100] -= 1
        for i in range(25):
            self.image = convolve(self.image, gaussMask, mode='same')
            self.image = 255 / num.amax(self.image) * self.image
            self.image[self.image>100] -= 1
        self.image = 80 / num.amax(self.image) * self.image
        self.image[self.image>40] += 40
        self.image[self.image>40] /= 2

    def randomPaintFilaments(self, rng):
        size = self.image.shape
        tmpImage = num.zeros(size)
        counter = 1
        print("Painting filaments:", end = " ")
        for filament in self.filaments:
            print(counter, end = " ")
            sys.stdout.flush()
            counter += 1
            reach = int(math.ceil(filament.width))
            gaussMask = makeGaussMask((filament.width - 0.3) / 1.5, reach)
            scale = 255. * filament.width
            pixels = filament.pixels()
            darkPixels = numberOfPixels = len(pixels)
            # Define canvas to paint this filament.
            halfWidth = int(abs(pixels[-1].x - pixels[-2].x) / 2 + 15)
            halfHeight = int(abs(pixels[-1].y - pixels[-2].y) / 2 + 15)
            cornerUpLeft = filament.center.asInt().shift(-halfWidth, -halfHeight)
            canvasSize = (2 * halfWidth, 2 * halfHeight)
            meanBrightness = rng()

            while darkPixels > 0.1 * numberOfPixels and meanBrightness < 255:
                darkPixels = 0
                filImage = num.zeros(canvasSize)
                filImageSmooth = num.zeros(canvasSize)
                for p in pixels:
                    filImage[p.x - cornerUpLeft.x, p.y - cornerUpLeft.y] = scale
                filImageSmooth = convolve(filImage, gaussMask, mode='same')
                filImageSmooth[filImageSmooth > 255] = 255
                brightnessMod = arandom.normal(meanBrightness/255., 0.1, filImageSmooth.shape)
                k = 30
                tmp = (filImageSmooth - k)
                tmp = brightnessMod * tmp + k
                filImageSmooth[filImageSmooth > k] = tmp[filImageSmooth > k]
                filImageSmooth[filImageSmooth > 255] = 255
                filImageSmooth = filImageSmooth.astype(num.uint8)
                for p in pixels:
                    if (filImageSmooth[p.x - cornerUpLeft.x, p.y - cornerUpLeft.y] <
                        self.image[p.x, p.y]):
                        darkPixels += 1
                meanBrightness += 5

            tmpImage = overlayImages(tmpImage, filImageSmooth, cornerUpLeft.x, cornerUpLeft.y)
        print('done!')
        tmpImage = overlayImages(self.image, tmpImage)
        sigma = 0.51
        reach = int(math.ceil(2 * sigma))
        gaussMask = makeGaussMask(sigma, reach)
        self.image = convolve(tmpImage, gaussMask, mode='same').astype(num.uint8)
        return self.image

    def noise (self, noise, scale):
        white_noise = noise(self.image.shape)
        self.image = (self.image.astype(num.float32) + white_noise) * scale()
        self.image[self.image > 255] = 255
        self.image[self.image < 0] = 0
        self.image = self.image.astype(num.uint8)
        return self.image


################################################################################
##############   Builder functions for random number generators   ##############
################################################################################

def makeMultiGaussRNG(centers, sigmas, weights):
    def rng():
        chooser = random.random()
        limit = 0
        out = None
        try:
            for i in range(len(weights)):
                limit += weights[i]
                if chooser < limit:
                    out = random.gauss(centers[i], sigmas[i]) % 180
            if out == None:
                out = random.gauss(centers[len(weights)], sigmas[len(weights)]) % 180
            if out < 0:
                out += 180
        except:
            print("RNG is broken.")
        return out
    return rng

def makeConstrainedRNG(minimum, maximum):
    def rng():
        return minimum + (maximum - minimum) * random.random()
    return rng

def makeConstrainedGaussianRNG(center, sigma, minimum, maximum):
    def rng():
        out = minimum - 1
        while (out < minimum or out > maximum):
            out = random.gauss(center, sigma)
        return out
    return rng

def makeMatrixGaussianRNG(center, sigma):
    def rng(shape):
        return arandom.normal(center, sigma, shape)
    return rng

def makeFilamentCenterRNG(ellipse):
    rng = makeConstrainedRNG(-ellipse.longAxis, ellipse.longAxis)
    def centerRNG():
        center = CENTER.copy().shift(rng(), rng())
        while not ellipse.includes(center):
            center = CENTER.copy().shift(rng(), rng())
        return center
    return centerRNG

def makePositionRNG(mask):
    x, y = mask.shape
    def rng():
        try_again = True
        while (try_again):
            position = (random.randint(0, x - 1), random.randint(0, y - 1))
            try_again = not mask[position]
        return position
    return rng

################################################################################
########################   Image processing functions   ########################
##################################################################4##############

def overlayImages(image1, image2, dx = 0, dy = 0):
    size1 = image1.shape
    size2 = image2.shape
    xHigh = max(size1[0], size2[0] + dx)
    yHigh = max(size1[1], size2[1] + dy)
    ext1 = num.zeros((xHigh, yHigh), dtype=image1.dtype)
    ext1[:size1[0], :size1[1]] = image1[:,:]
    ext2 = num.zeros((xHigh, yHigh), dtype=image2.dtype)
    ext2[dx:(size2[0] + dx), dy:(size2[1] + dy)] = image2[:,:]
    image = num.maximum(ext1, ext2)
    return image

def makeGaussMask(sigma, size):
    try:
        sumGauss = 0
        mask = num.zeros((2*size+1, 2*size+1))
        for a in range(-size, size + 1):
            for b in range(-size, size + 1):
                mask[size+a, size+b] = math.exp(-(a**2 + b**2) / (2.0 * sigma))
                sumGauss += mask[size+a, size+b]
        return mask / sumGauss
    except:
        print("Could not make Gaussian mask.")
    return None

# This function exists to compensate for numpy's weird coordinate system.
def toImage(array):
    return img.fromarray(num.swapaxes(array, 0, 1))

# This function assembles a cell image using random number generators and the
# methods of the Cell class.
def makeRandomCell(name):
    # Make ellipse as basis for cell area.
    axisRNG = makeConstrainedGaussianRNG(130, 80, 50, 300)
    cellOrientationRNG = makeConstrainedRNG(0, 180)
    ellipse = Ellipse.random(axisRNG, axisRNG, cellOrientationRNG)

    # Make filaments
    n = ellipse.area() / 1000
    numberOfFilaments = max(25, int(makeConstrainedGaussianRNG(n, 15, 0.7 * n, 1.3 * n)()))
    centerRNG = makeFilamentCenterRNG(ellipse)
    lengthRNG = makeConstrainedGaussianRNG(100, 50, 25, 200)
    widthRNG = makeConstrainedGaussianRNG(0, 1.6, 1, 5)

    numberOfModes = int(makeConstrainedGaussianRNG(0, 2, 1, 2.9)())
    sigmaRNG = makeConstrainedGaussianRNG(15, 8, 5, 30)
    centers = []
    sigmas = []
    weights = []
    for i in range(numberOfModes):
        centers.append(180*random.random())
        sigmas.append(sigmaRNG())
        if i == 1:
            weights.append(random.random())
        elif i > 1:
            weights.append((1 - sum(weights)) * random.random())
    orientationRNG = makeMultiGaussRNG(centers, sigmas, weights)

    filaments = []
    csv_output = open(("%s/%s.csv" % (OUTPUT_FOLDER, name)), "w")
    csv_output.write("#center_x,center_y,length,angle,width\n")
    for i in range(numberOfFilaments):
        fil = Filament.random(centerRNG, lengthRNG, widthRNG, orientationRNG)
        if fil.length > 75 and random.randint(0,2):
            twin = Filament.twin(fil)
            filaments += [fil, twin]
            csv_output.write("%s\n%s\n" % (fil.out(), twin.out()))
        else:
            filaments += [fil]
            csv_output.write("%s\n" % fil.out())
    csv_output.close()

    # Paint cell
    cell = Cell(ellipse, filaments)
    cell.background(0.08)
    toImage(cell.image).show()

    filamentBrightnessRNG = makeConstrainedGaussianRNG(90, 40, 60, 255)
    cell.randomPaintFilaments(filamentBrightnessRNG)
    noiseRNG = makeMatrixGaussianRNG(0, 5)
    scaleRNG = makeConstrainedGaussianRNG(1.0, 0.2, 1.0, 1.5)
    image = toImage(cell.noise(noiseRNG, scaleRNG))
    image.show()
    image.save("%s/%s.png" % (OUTPUT_FOLDER, name))

# Call the control function.
if __name__ == '__main__':
    makeRandomCell("test_case_cell")