# -*- coding: utf-8 -*-
"""
Copyright (C) 2014 Benjamin Eltzner

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
or see <http://www.gnu.org/licenses/>.
"""

import os, math
import numpy as num
import matplotlib.pyplot as plot
import multiprocessing as mp
from subprocess import Popen, PIPE
from scipy.misc import imread
from PIL import Image as img
from rpy2.robjects.packages import importr as r_import
from rpy2.robjects import r as r_input

PI = num.pi
RAD = math.radians(1)
SIGMA = 10.0
LABEL_SIZE = 18.0
OUTPUT_DIRECTORY = 'output/'

"""
AUTHOR'S DISCLAIMER:
This software was written in view of a specific use case on a specific system
and is provided for the sake of verifiability of its results. It is not
recommended to reuse this software in any other scenario.

The software makes use of output functions without checking for pre-existing
data. Running it CAN RESULT IN LOSS OF DATA by unchecked overwriting and
CAN POTENTIALLY DAMAGE YOUR SYSTEM.

This disclaimer need not be reproduced in copies or derived works of this
software. However, the author would be grateful if it was kept intact at least
in unmodified copies.
"""

################################################################################
##############################   Input functions   #############################
################################################################################

def importFile (filename, withNames = False):
    data = []; names = []
    with open(filename) as datafile:
        for line in datafile:
            if line[0] != '#' and line[0] != '%':
                dataLine = line.strip().split(',')
                if withNames:
                    data.append([float(x) for x in dataLine[1:]])
                    names.append(dataLine[0])
                else:
                    data.append([float(x) for x in dataLine])
    if withNames:
        return names, data
    return num.array(data)

def importBinaryImage (filename):
    binaryImage = []
    with open(filename) as datafile:
        for line in datafile:
            binaryImage.append([bool(int(item)) for item in line.strip().split(' ')])
    binaryImage = num.array(binaryImage, dtype=bool).T
    #toImage(binaryImage.astype(num.uint8)*255).show()
    return binaryImage

def imageToArray (filename):
    #toImage(binaryImage.astype(num.uint8)*255).show()
    return num.array(num.sum(imread(filename)[:,:,:3], axis=2) != 0, dtype=bool).T

def importImage (filename, scale = True):
    image = []
    with open(filename) as datafile:
        for line in datafile:
            image.append([float(item) for item in line.strip().split(' ')])
    image = num.array(image).T
    print filename, image.shape, num.amax(image)
    if scale:
        image = image * 255. / num.amax(image)
        image[image < 0] = 0
    else:
        image = ((image - num.amax(image)) * 255. /
                 (num.amax(image) - num.amin(image)))
    image = num.array(image, dtype=num.uint8)
    #toImage(image).show()
    return image

# This function is specific to unixoid systems.
def getFiles (dataType, ending = '_*.csv'):
    pipe = Popen(['find', '.', '-wholename', '*' + dataType + ending], stdout = PIPE)
    output, _ = pipe.communicate()
    output = output.strip().split('\n')
    return sorted(output, key = str.lower)

def getNames (paths):
    fileNames = []
    for path in paths:
        fileNames.append(os.path.split(path)[1])
    return fileNames

################################################################################
##############################   Plot functions   ##############################
################################################################################

"""
All plot functions use the program sed present on unixoid systems to remove the
reference to page groups from the generated pdf files. This is necessary to
avoid this bug in adobe reader:
https://tex.stackexchange.com/questions/76273/multiple-pdfs-with-page-group-included-in-a-single-page-warning

As pyplot keeps a reference to every plot it has generated, which effectively
implies a memory leak, all plotting is wrapped in a worker thread that is
disposed (and thus frees memory) after plotting.
"""

def makeBarChart (data, fileName = None, identifier = '', title = None):
    fig = plot.figure()
    if title != None:
        fig.suptitle(title, fontsize = 10)
    diag = fig.add_subplot(111)
    data = num.concatenate((data, num.array([[180., data[0,1]]])))
    xdata = data[:, 0] - (90 / float(len(data) - 1))
    ydata = data[:, 1]
    line = histogram2gauss(data)
    binWidth = 180 / float(len(data) - 1)
    diag.bar(xdata, ydata, width=binWidth, linewidth=0, color='#888888')
    diag.plot(line[0], line[1], linewidth=3.0, color='black')
    diag.set_xlim(0,180)
    diag.set_ylim(bottom = 0)
    diag.set_xlabel('Orientation [${}^\circ$]', fontsize=LABEL_SIZE)
    diag.set_ylabel('Pixels', fontsize=LABEL_SIZE)
    plot.tick_params(labelsize=LABEL_SIZE)
    plot.tight_layout()
    if (not fileName == None and fileName.endswith('.csv')):
        fig.savefig(fileName[:-4] + '_' + identifier + '.tif', dpi= 300)
    else:
        fig.show()
    plot.close()

def makePixelPlot (data, ticks, x_label, y_label, name, fileName = None, log_axis = False):
    data = num.array(data)
    fig = plot.figure()
    diag = fig.add_subplot(111)
    xdata = range(1,len(data)+1)
    plot.xticks(range(1,len(ticks)+1), ticks)
    diag.plot(xdata, data[:,0], color='#888888')
    diag.plot(xdata, data[:,1], color='#444444')
    diag.plot(xdata, data[:,2], color='#000000')
    p0 = diag.scatter(xdata, data[:,0], color='#888888', marker='^', s = 50)
    p1 = diag.scatter(xdata, data[:,1], color='#444444', marker='*', s = 50)
    p2 = diag.scatter(xdata, data[:,2], color='#000000', marker='o', s = 50)
    diag.set_xlim(0,11)
    diag.set_xlabel(x_label, fontsize=LABEL_SIZE-5)
    diag.set_ylim(bottom = 0)
    diag.set_ylabel(y_label, fontsize=LABEL_SIZE-5)
    if log_axis:
        diag.set_yscale('log')
        bottom = 1
        while bottom > num.amin(data):
            bottom *= 0.1
        diag.set_ylim(bottom = bottom)
    diag.legend([p0,p1,p2], (name, 'eLoG', 'FS'), loc ='best', scatterpoints = 1)
    plot.tick_params(labelsize=LABEL_SIZE)
    _, labels = plot.xticks()
    plot.setp(labels, rotation=45, fontsize=LABEL_SIZE-5)
    plot.tight_layout()
    if (not fileName == None):
        plot.savefig(OUTPUT_DIRECTORY + fileName + '.tif', dpi= 300, compress='lzw')
    else:
        plot.show()
    plot.close()

def makeBoxPlot (data, label, name, fileName = None, log_axis = False):
    data = num.array(data)
    fig = plot.figure()
    diag = fig.add_subplot(111)
    boxplot = diag.boxplot(data, patch_artist=True, sym='bo')
    diag.set_ylim(bottom = 0)
    plot.xticks([1, 2, 3], [name, 'eLoG', 'FS'])
    plot.tick_params(labelsize=LABEL_SIZE)
    diag.set_xlabel('Method', fontsize=LABEL_SIZE-5)
    diag.set_ylabel(label, fontsize=LABEL_SIZE-5)
    plot.setp(boxplot['boxes'], color='black', facecolor='#CCCCCC', linewidth=2)
    plot.setp(boxplot['whiskers'], color='black', linestyle='-', linewidth=2)
    plot.setp(boxplot['caps'], color='black', linestyle='-', linewidth=2)
    plot.setp(boxplot['fliers'], markerfacecolor='black')
    plot.setp(boxplot['medians'], color='black', linestyle='-', linewidth=2)
    if log_axis:
        diag.set_yscale('log')
        bottom = 1
        while bottom > num.amin(data):
            bottom *= 0.1
        diag.set_ylim(bottom = bottom)
    plot.tight_layout()
    if (not fileName == None):
        plot.savefig(OUTPUT_DIRECTORY + fileName + '.tif', dpi= 300, compress='lzw')
    else:
        plot.show()
    plot.close()

def makeErrorPlot (data, label, name, fileName = None):
    data = num.array(data)
    fig = plot.figure()
    diag = fig.add_subplot(111)
    means = num.mean(data, axis=0)
    standevs = num.std(data, axis=0)
    diag.errorbar([1,2,3], means, yerr=standevs, color='#000000', marker='o',
                  ms=15, fmt='o', elinewidth=2, capthick=2, capsize=10)
    diag.set_xlim(0.5, 3.5)
    diag.set_ylim(bottom = 0)
    plot.xticks([1, 2, 3], [name, 'eLoG', 'FS'])
    plot.tick_params(labelsize=LABEL_SIZE)
    diag.set_xlabel('Method', fontsize=LABEL_SIZE-5)
    diag.set_ylabel(label, fontsize=LABEL_SIZE-5)
    plot.tight_layout()
    if (not fileName == None):
        plot.savefig(OUTPUT_DIRECTORY + fileName + '.tif', dpi= 300, compress='lzw')
    else:
        plot.show()
    plot.close()

# Worker thread to encapsulate plotting and avoid pyplot memory leak.
def plotThread (worker):
    proc = mp.Process(target = worker)
    proc.daemon = True
    proc.start()
    proc.join()

################################################################################
######################   Histogram processing functions   ######################
################################################################################

# For Hough data
def recalculateOrientations (data, numberOfOrientations = None):
    if numberOfOrientations == None:
        numberOfOrientations = len(data)
    for item in data:
        item[0] *= 180 / float(numberOfOrientations)
    return data

# For filament sensor and ground truth data
def makeBinning (data):
    orientations = data[:, 3]
    amplitudes = num.multiply(data[:, 2], data[:, 4])
    binned = num.zeros(180)
    for i in range(len(orientations)):
        #print orientations[i],
        binned[round(orientations[i]) % 180] += amplitudes[i]
    binned = (num.vstack([num.array(range(180)), binned])).T
    return binned

def removeEmptyBins(histogram):
    listEmpty = []
    for i in range(len(histogram)):
        if histogram[i,1] == 0:
            listEmpty.append(i)
    return num.delete(histogram, listEmpty, axis=0)

def histogram2gauss (data):
    steps = 10
    stepWidth = 1 / float(steps)
    gauss = gaussian(SIGMA, stepWidth)
    length = len(gauss)
    shift = int((length - 1) / 2)
    x_values = num.arange(0, len(data) + 0.001, stepWidth) * abs(data[1,0] - data[2,0])
    tmp = num.zeros(len(data) * steps + length)
    for i, [_, amplitude] in enumerate(data):
        try:
            tmp[i*steps: length+i*steps] += amplitude * gauss
        except:
            print len(gauss), len(tmp[i*steps: length+i*steps])
    y_values = tmp[shift:-shift]
    y_values[:shift] += tmp[-shift:]
    y_values[-shift:] += tmp[:shift]
    return num.vstack((x_values, y_values))

def gaussian (sigma, step):
    x_values = num.arange(-5 * sigma, 5 * sigma + 0.001, step)
    return num.exp(-x_values**2 / (2 * sigma**2)) / (math.sqrt(2 * PI) * sigma)

################################################################################
##########################   Bookkeeping functions   ###########################
################################################################################

def getListsForBinaryImages (imageType):
    files = (getFiles(imageType) + getFiles(imageType + '*bin_img', '') +
             getFiles(imageType + '*eLoG', '*filaments.png'))
    fileLists = [['truth'], ['sensor'], ['cid'], ['elog']]
    fileNames = []
    for f in files:
        if 'truth' in f:
            fileLists[0].append(f)
            fileNames.append(getNames([f])[0][:-4])
        if 'sensor' in f:
            fileLists[1].append(f)
        if 'CID' in f:
            fileLists[2].append(f)
        if 'eLoG' in f and '.png' in f:
            fileLists[3].append(f)
    return fileNames, fileLists

def getShape (name, fileLists):
    shape = None
    fileDictionary = {'filename' : name}
    for fileList in fileLists:
        if fileList[0] == 'cid':
            #print fileList[0]
            found = False
            i = 1
            while not found:
                if i >= len(fileList):
                    print name, 'not found in', fileList[0]
                    return None
                if name in fileList[i]:
                    fileDictionary[fileList[0]] = importBinaryImage(fileList[i])
                    shape = fileDictionary[fileList[0]].shape
                    found = True
                i += 1
    return shape, fileDictionary

def listBinaryImages (imageType = ''):
    print imageType
    fileNames, fileLists = getListsForBinaryImages(imageType)
    print fileNames
    listDictionaries = []
    for name in fileNames:
        #print 'Make dictionary for', name
        shape, fileDictionary = getShape(name, fileLists)
        for fileList in fileLists:
            if fileList[0] != 'cid':
                #print fileList[0]
                found = False
                i = 1
                while not found:
                    if i >= len(fileList):
                        print name, 'not found in', fileList[0]
                        return None
                    if name in fileList[i]:
                        if fileList[0] == 'truth':
                            fileDictionary[fileList[0]] = makeBinaryImage(importFile(fileList[i]), shape)
                        elif fileList[0] == 'sensor':
                            fileDictionary[fileList[0]] = makeBinaryImage(importFile(fileList[i]), shape)
                        elif fileList[0] == 'elog':
                            fileDictionary[fileList[0]] = imageToArray(fileList[i])
                        else:
                            fileDictionary[fileList[0]] = importBinaryImage(fileList[i])
                        found = True
                    i += 1
        listDictionaries.append(fileDictionary)
    return sorted(listDictionaries, key=lambda d: d["filename"])

def makeDictionary (name, invert, *args):
    fileDictionary = {'filename' : name}
    for fileList in args:
        found = False
        i = 1
        while not found:
            if i >= len(fileList):
                print name, 'not found in', fileList[0]
                return None
            if name[:-4] in fileList[i]:
                if fileList[0] == 'sensor':
                    fileDictionary[fileList[0]] = removeEmptyBins(makeBinning(importFile(fileList[i])))
                elif fileList[0] == 'truth':
                    fileDictionary[fileList[0]] = removeEmptyBins(makeBinning(importFile(fileList[i])))
                elif fileList[0] == 'hough':
                    fileDictionary[fileList[0]] = removeEmptyBins(recalculateOrientations(importFile(fileList[i])))
                else:
                    fileDictionary[fileList[0]] = removeEmptyBins(importFile(fileList[i]))
                found = True
            i += 1
    return fileDictionary

def listDictionaries (imageType = '', invert = False):
    files = getFiles(imageType)
    print imageType #, '\n', files
    filesTruth = ['truth']
    filesSensor = ['sensor']
    filesElog = ['elog']
    filesHough = ['hough']
    for f in files:
        if 'truth' in f:
            filesTruth.append(f)
        if 'sensor' in f:
            filesSensor.append(f)
        if 'eLoG' in f:
            filesElog.append(f)
        if 'hough' in f:
            filesHough.append(f)
    fileNames = getNames(filesTruth[1:])
    listDictionaries = []
    for name in fileNames:
        listDictionaries.append(makeDictionary(name, invert, filesTruth,
                                               filesSensor, filesElog, filesHough))
    return sorted(listDictionaries, key=lambda d: d["filename"])

# This method exists to compensate for numpy's weird coordinate system.
def toImage(array):
    return img.fromarray(num.swapaxes(array, 0, 1))

################################################################################
#########################   Binary image comparison   ##########################
################################################################################

def onLine(point, line):
    if line[4] == 0: return False
    t = num.dot((point - line[:2]), line[2:4]) / line[4]
    t = (0 if t < 0 else (1 if t > 1 else t))
    dist = (num.linalg.norm(point - line[:2] - t * line[2:4]))
    return dist <= line[5]

def makeBinaryImage (data, shape):
    binaryImage = num.zeros(shape, dtype = bool)
    for i in range(len(data)):
        line = num.zeros(6)
        # [headx, heady, tailx - headx, taily - heady, lengthSquare, halfWidth]
        line[:2] = [round(data[i,0] - data[i,2] * math.cos(data[i,3] * RAD) / 2.0),
                    round(data[i,1] + data[i,2] * math.sin(data[i,3] * RAD) / 2.0)]
        line[2:4] = [round(data[i,0] + data[i,2] * math.cos(data[i,3] * RAD) / 2.0) - line[0],
                     round(data[i,1] - data[i,2] * math.sin(data[i,3] * RAD) / 2.0) - line[1]]
        line[4:] = [num.dot(line[2:4],line[2:4]), (0.5 * data[i,4])]
        xMin = int(max(0, min(line[0], line[0] + line[2]) - math.ceil(line[5]) - 1))
        xMax = int(min(shape[0], max(line[0], line[0] + line[2]) + math.ceil(line[5]) + 1))
        yMin = int(max(0, min(line[1], line[1] + line[3]) - math.ceil(line[5]) - 1))
        yMax = int(min(shape[1], max(line[1], line[1] + line[3]) + math.ceil(line[5]) + 1))
        for x in range(xMin, xMax):
            for y in range (yMin, yMax):
                binaryImage[x,y] = binaryImage[x,y] or onLine([x,y], line)
    #toImage(binaryImage.astype(num.uint8)*255).show()
    return binaryImage

def compareBinaryImages (image1, image2, tol):
    shape = image1.shape
    matches = num.zeros(3)
    misses = num.zeros(2)
    posNegImg = num.zeros(shape)
    for x in range(shape[0]):
        for y in range(shape[1]):
            if not image1[x,y] and not image2[x,y]:
                pass
            elif image1[x,y]:
                if image2[x,y]:
                    matches[0] += 1
                elif (x in range(tol, shape[0] - tol) and
                      y in range(tol, shape[0] - tol) and
                      image2[x-tol:x+tol+1,y-tol:y+tol+1].any()):
                    matches[1] += 1
                else:
                    misses[0] += 1
                    posNegImg[x,y] = 100 # false negative
            elif image2[x,y]:
                if (x in range(tol, shape[0] - tol) and
                    y in range(tol, shape[1] - tol) and
                    image1[x-tol:x+tol+1,y-tol:y+tol+1].any()):
                    matches[2] += 1
                else:
                    misses[1] += 1
                    posNegImg[x,y] = 200 # false positive
            else:
                print 'Why am I here\nand for how long?'
    return matches, misses, posNegImg

################################################################################
##########################   Earth movers distance   ###########################
################################################################################

def angularMetric ():
    def metric(x,y):
        return min(abs(x-y), 180 - abs(x-y))
    return metric

def normalize (table):
#    print 'normalize\n', table
    scale = sum(table[:,1])
    out = num.zeros(table.shape)
    out[:,0] = table[:,0]
    out[:,1] = table[:,1] / float(scale)
    return out, scale

def costMatrix (source, target, metric):
    costMatrix = num.zeros((len(source), len(target)))
    for x in range(len(source)):
        for y in range(len(target)):
            costMatrix[x,y] = metric(source[x,0],target[y,0])
    return costMatrix

def r_vec (array, name):
    r_input(name + '<-c(' + ','.join(str(i) for i in array) + ');')

def r_mat (matrix, name):
    r_input(name + '<-array(c(' + ','.join(str(i) for i in matrix.T.flatten()) +
            '), dim=c' + str(matrix.shape) + ');')

def r_transport (start, target, cost):
    r_vec(start, 'start')
    r_vec(target, 'target')
    r_mat(cost, 'cost')
    r_input('result<-transport(start, target, cost);')
    sparse_quantities = num.array(r_input('result')).T
    quantities = num.zeros(cost.shape)
    for line in sparse_quantities:
        quantities[int(line[0] - 1), int(line[1] - 1)] = line[2]
    return num.einsum('ij,ij->', cost, quantities)

def calculateDistances (dictionary):
    metric = angularMetric()
    for key in dictionary.keys():
        if key != 'truth' and key != 'filename':
            source, source_scale = normalize(dictionary['truth'])
            target, target_scale = normalize(dictionary[key])
            matrix = costMatrix(source, target, metric)
            print 'Transporting', key
            dictionary[key] = [target_scale / source_scale,
                               r_transport(source[:,1], target[:,1], matrix)]
    del dictionary['truth']
    return dictionary

################################################################################
############################   Control functions   #############################
################################################################################

def drawIt ():
    files = getFiles('truth')
    for f in files:
        print 'Making chart for', f
        def worker ():
            makeBarChart(makeBinning(importFile(f)), f, 'ground_truth')
        plotThread(worker)
    files = getFiles('sensor')
    for f in files:
        print 'Making chart for', f
        def worker():
            makeBarChart(makeBinning(importFile(f)), f, 'line_sensor')
        plotThread(worker)
    files = getFiles('hough')
    for f in files:
        print 'Making chart for', f
        def worker():
            makeBarChart(recalculateOrientations(importFile(f)), f, 'hough')
        plotThread(worker)
    files = getFiles('eLoG')
    for f in files:
        print 'Making chart for', f
        def worker():
            makeBarChart(importFile(f), f, 'eLoG')
        plotThread(worker)
    print "It is done."

def moveIt ():
    r_import('transport')
    dataScaleReal = []
    dataScaleSim = []
    dataDistReal = []
    dataDistSim = []
    namesReal = []
    namesSim = []
    for imageType in ['ugly', 'bad', 'medium', 'good', 'simulated']:
        dictionaries = listDictionaries(imageType)
        fileScales = open(OUTPUT_DIRECTORY + imageType + '_scales.csv', 'w')
        fileDist = open(OUTPUT_DIRECTORY + imageType + '_distances.csv', 'w')
        fileScales.write('#cell,hough,elog,sensor\n')
        fileDist.write('#cell,hough,elog,sensor\n')
        for d in dictionaries:
            print 'Processing data for', d['filename']
            d = calculateDistances(d)
            print d
            fileScales.write('%s,%.4f,%.4f,%.4f\n' % (d['filename'], d['hough'][0],
                                                      d['elog'][0], d['sensor'][0]))
            fileDist.write('%s,%.4f,%.4f,%.4f\n' % (d['filename'], d['hough'][1],
                                                    d['elog'][1], d['sensor'][1]))
            output = open((OUTPUT_DIRECTORY + d['filename'][:-4] + '_table'), "w")
            for key in d.keys():
                if key != 'filename':
                    output.write("%s\n%.4f\n%.4f\n" % (key, d[key][0], d[key][1]))
            output.close()
            if imageType == 'simulated':
                dataScaleSim.append([d['hough'][0], d['elog'][0], d['sensor'][0]])
                dataDistSim.append([d['hough'][1], d['elog'][1], d['sensor'][1]])
                namesSim.append(''.join(i for i in d['filename'] if i.isdigit()))
            else:
                dataScaleReal.append([d['hough'][0], d['elog'][0], d['sensor'][0]])
                dataDistReal.append([d['hough'][1], d['elog'][1], d['sensor'][1]])
                namesReal.append(d['filename'][:-4])
        fileScales.close()
        fileDist.close()
    makeBoxPlot(dataScaleReal, 'Scale (log axis)', 'Hough', 'scales_real_error', True)
    makeBoxPlot(dataDistReal, 'Distance [${}^\circ$] (log axis)', 'Hough', 'dist_real_error', True)
    makeBoxPlot(dataScaleSim, 'Scale', 'Hough', 'scales_sim_error')
    makeBoxPlot(dataDistSim, 'Distance [${}^\circ$]', 'Hough', 'dist_sim_error')
    makePixelPlot(dataScaleReal, namesReal, 'Cell', 'Scale (log axis)', 'Hough', 'scales_real', True)
    makePixelPlot(dataDistReal, namesReal, 'Cell', 'Distance [${}^\circ$] (log axis)', 'Hough', 'dist_real', True)
    makePixelPlot(dataScaleSim, namesSim, 'Simulated cell', 'Scale', 'Hough', 'scales_sim')
    makePixelPlot(dataDistSim, namesSim, 'Simulated cell', 'Distance [${}^\circ$]', 'Hough', 'dist_sim')
    print "That's it, folks!"

def countIt():
    dataFalseNegativeReal = []; dataFalseNegativeSim = []
    dataFalsePositiveReal = []; dataFalsePositiveSim = []
    namesReal = [];             namesSim = []
    for imageType in ['ugly', 'bad', 'medium', 'good', 'simulated']:
        dictionaries = listBinaryImages(imageType)
        fileResult = open(OUTPUT_DIRECTORY + imageType + '_pixels.csv', 'w')
        fileResult.write('#cell,cid_hit,cid_neg,cid_pos,elog_hit,elog_neg,elog_pos,sensor_hit,sensor_neg,sensor_pos\n')
        for d in dictionaries:
            print 'Processing data for', d['filename']
            hitCid, missCid, imgCid = compareBinaryImages(d['truth'],d['cid'],1)
            hitElog, missElog, imgElog = compareBinaryImages(d['truth'],d['elog'],1)
            hitSensor, missSensor, imgSensor = compareBinaryImages(d['truth'],d['sensor'],1)
            toImage(imgCid.astype(num.uint8)).save(OUTPUT_DIRECTORY + "%s_%s_CID.png" % (d['filename'], imageType))
            toImage(imgElog.astype(num.uint8)).save(OUTPUT_DIRECTORY + "%s_%s_eLoG.png" % (d['filename'], imageType))
            toImage(imgSensor.astype(num.uint8)).save(OUTPUT_DIRECTORY + "%s_%s_Sensor.png" % (d['filename'], imageType))
            print hitCid, missCid, '\n', hitElog, missElog, '\n', hitSensor, missSensor
            values = tuple([d['filename'], sum(hitCid[:2])] + list(missCid) +
                           [sum(hitElog[:2])] + list(missElog) +
                           [sum(hitSensor[:2])] + list(missSensor))
            fileResult.write('%s,%d,%d,%d,%d,%d,%d,%d,%d,%d\n' % values)
            sumCid = float(sum(hitCid[:2]) + missCid[0])
            sumElog = float(sum(hitElog[:2]) + missElog[0])
            sumSensor = float(sum(hitSensor[:2]) + missSensor[0])
            if imageType == 'simulated':
                dataFalseNegativeSim.append([missCid[0]/sumCid, missElog[0]/sumElog, missSensor[0]/sumSensor])
                dataFalsePositiveSim.append([missCid[1]/sumCid, missElog[1]/sumElog, missSensor[1]/sumSensor])
                namesSim.append(''.join(i for i in d['filename'] if i.isdigit()))
            else:
                dataFalseNegativeReal.append([missCid[0]/sumCid, missElog[0]/sumElog, missSensor[0]/sumSensor])
                dataFalsePositiveReal.append([missCid[1]/sumCid, missElog[1]/sumElog, missSensor[1]/sumSensor])
                namesReal.append(d['filename'])
        fileResult.close()
    makeBoxPlot(dataFalseNegativeReal, 'False negative rate (log axis)', 'CID', 'false_neg_real_box', True)
    makeBoxPlot(dataFalsePositiveReal, 'False positive rate (log axis)', 'CID', 'false_pos_real_box', True)
    makeBoxPlot(dataFalseNegativeSim, 'False negative rate', 'CID', 'false_neg_sim_box')
    makeBoxPlot(dataFalsePositiveSim, 'False positive rate', 'CID', 'false_pos_sim_box')
    makePixelPlot(dataFalseNegativeReal, namesReal, 'Cell', 'False negative rate (log axis)', 'CID', 'false_neg_real', True)
    makePixelPlot(dataFalsePositiveReal, namesReal, 'Cell', 'False positive rate (log axis)', 'CID', 'false_pos_real', True)
    makePixelPlot(dataFalseNegativeSim, namesSim, 'Simulated cell', 'False negative rate', 'CID', 'false_neg_sim')
    makePixelPlot(dataFalsePositiveSim, namesSim, 'Simulated cell', 'False positive rate', 'CID', 'false_pos_sim')
    print "Fin"

def pngIt ():
    for imageType in ['*bin_img', '*xcorr']:
        files = getFiles(imageType, '')
        for f in files:
            if imageType == '*bin_img':
                img = toImage(importBinaryImage(f).astype(num.uint8)*255)
            if imageType == '*xcorr':
                img = toImage(importImage(f, True))
                img.show()
                img.save("%s_scaled.png" % f)
                img = toImage(importImage(f, False))
            img.show()
            img.save("%s.png" % f)

def replot():
    for name in ['scales', 'distances', 'pixels']:
        files = getFiles(OUTPUT_DIRECTORY + '*' + name, '*.csv')
        print name, files
        dataReal = []; namesReal = []
        for imageType in ['simulated']:
            [(_, dataSim)] = [importFile(f, True) for f in files if imageType in f]
        for imageType in ['ugly', 'bad', 'medium', 'good']:
            [(ntmp, dtmp)] = [importFile(f, True) for f in files if imageType in f]
            dataReal += dtmp
            namesReal += ntmp
        dataSim = num.array(dataSim); dataReal = num.array(dataReal)
        if name == 'scales':
            makeBoxPlot(dataReal, 'Scale (log axis)', 'Hough', 'scales_real_box', True)
            makeBoxPlot(dataSim, 'Scale', 'Hough', 'scales_sim_box')
        if name == 'distances':
            makeBoxPlot(dataReal, 'Distance [${}^\circ$] (log axis)', 'Hough', 'dist_real_box', True)
            makeBoxPlot(dataSim, 'Distance [${}^\circ$]', 'Hough', 'dist_sim_box')
        if name == 'pixels':
            for i, data in enumerate([dataReal, dataSim]):
                sumCid = num.sum(data[:,0:2], axis=1)
                sumElog = num.sum(data[:,3:5], axis=1)
                sumSensor = num.sum(data[:,6:8], axis=1)
                if i == 1:
                    dataFalseNegativeSim = num.vstack((data[:,1]/sumCid, data[:,4]/sumElog, data[:,7]/sumSensor)).T
                    dataFalsePositiveSim = num.vstack((data[:,2]/sumCid, data[:,5]/sumElog, data[:,8]/sumSensor)).T
                else:
                    dataFalseNegativeReal = num.vstack((data[:,1]/sumCid, data[:,4]/sumElog, data[:,7]/sumSensor)).T
                    dataFalsePositiveReal = num.vstack((data[:,2]/sumCid, data[:,5]/sumElog, data[:,8]/sumSensor)).T
            makeBoxPlot(dataFalseNegativeReal, 'False negative rate (log axis)', 'CID', 'false_neg_real_box', True)
            makeBoxPlot(dataFalsePositiveReal, 'False positive rate (log axis)', 'CID', 'false_pos_real_box', True)
            makeBoxPlot(dataFalseNegativeSim, 'False negative rate', 'CID', 'false_neg_sim_box')
            makeBoxPlot(dataFalsePositiveSim, 'False positive rate', 'CID', 'false_pos_sim_box')


# Call the control functions.
if __name__ == "__main__":
    #drawIt()
    #moveIt()
    #countIt()
    #pngIt()
    replot()
