package ui;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.io.File;
import java.io.FileWriter;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;

import javax.xml.transform.dom.DOMSource;

import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * FilamentSensor configuration class. This class implements the configuration
 * interface and loading/saving the config to XML. The main configuration covers
 * window size, position and all persistent GUI variables.
 */
public abstract class Config
{
    // Default values
    private static final Map<String, String> defaults = mapDefaults();

    private static Map<String, String> mapDefaults()
    {
        Map<String, String> map = new HashMap<String, String>();
        map.put("FilamentSensor.show_intro", "1");
        map.put("MainWnd.size_w", "800");
        map.put("MainWnd.size_h", "640");
        map.put("tabBar.divider_location", "400");

        return Collections.unmodifiableMap(map);
    }

    // Class variables
    private static DocumentBuilder m_document_builder;
    private static String m_configuration_file_path;
    private static Document m_config_document;
    private static File m_database;
    private static final String[] OLD_CONFIGS = new String[]{".filtracker"};

    // Integer constants
    public static final int PROJECT = 0, IMAGE = 1;
    private static final Map<Integer, String> m_type_map = mapTypes();
    
    private static Map<Integer, String> mapTypes()
    {
        Map<Integer, String> m = new HashMap<Integer, String>();
        m.put(PROJECT, "Project_Files");
        m.put(IMAGE, "Image_Data_Files");

        return Collections.unmodifiableMap(m);
    }

    public static void init(String config_folder)
    {
        // Determine home-directory
        String home = System.getProperty("user.home");
        if (home == null && FilamentSensor.DEBUG)
        {
            System.out.println("Warning: could not determine home directory");
        }
        // File seperator, e.g. '/' for Unixoid, '\' for Windows
        String fspr = System.getProperty("file.separator");

        m_configuration_file_path = "";
        if (home != null)
        {
            File path = new File(home + fspr + config_folder);
            if (path.mkdirs() && FilamentSensor.DEBUG)
            {
                System.out.println("Folder '" + home + fspr + config_folder + "' created.");
            }

            m_configuration_file_path = home + fspr + config_folder + fspr + "config.xml";
            m_database = new File(home + fspr + config_folder + fspr + "database.xml");
        }
        
        try
        {
            m_document_builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        }
        catch (ParserConfigurationException e)
        {
            System.out.println("Error while initializing document builder: " + e.getMessage());
            return;
        }

        if (!m_database.exists() || !m_database.canRead() || !m_database.isFile())
        {
            boolean read = false;
            for (String path : OLD_CONFIGS)
            {
                File database = new File(home + fspr + path + fspr + "database.xml");
                if (database.exists() && database.canRead() && database.isFile())
                {
                    read = true;
                    database.renameTo(m_database);
                    renameDatabaseTag();
                }
            }

            if (!read)
            {
                if (FilamentSensor.DEBUG)
                {
                    System.out.println("No valid database found at " + m_database
                                       + ". Creating new one.");
                }

                Document database_document = m_document_builder.newDocument();
                Element root = database_document.createElement("FilamentSensorDatabase");
                database_document.appendChild(root);

                for (String type : m_type_map.values())
                {
                    root.appendChild(database_document.createElement(type));
                }

                try
                {
                    // write the created document object to file
                    Transformer transformer = TransformerFactory.newInstance().newTransformer();
                    transformer.setOutputProperty(OutputKeys.INDENT, "yes");

                    FileWriter writer = new FileWriter(m_database);
                    transformer.transform(new DOMSource(database_document), new StreamResult(writer));
                    writer.close();
                }
                catch (Exception e) // IOException or TransformerConfigurationException
                {
                    System.out.println("Error while initializing '" + m_database + "': " + e.getMessage());
                }
            }
        }
        else
        {
            renameDatabaseTag();
        }

        // Initialize doc
        if (!readConfig(m_configuration_file_path))
        {
            boolean read = false;
            for (String path : OLD_CONFIGS)
            {
                path = home + fspr + path + fspr + "config.xml";
                if(readConfig(path))
                {
                    read = true;
                    saveConfig();
                    (new File(path)).delete();
                }
            }

            if (!read)
            {
                m_config_document = m_document_builder.newDocument();
                Element root = m_config_document.createElement("FilamentSensor");
                m_config_document.appendChild(root);

                if (FilamentSensor.DEBUG)
                {
                    System.out.println("New config-Document created.");
                }
            }
        }

        for (String path : OLD_CONFIGS)
        {
            File folder = new File(home + fspr + path);
            if(folder.exists() && folder.isDirectory() && folder.list().length <= 0)
            {
                folder.delete();
            }
        }
    }

    private static boolean renameDatabaseTag()
    {
        try
        {
            Document database_document = m_document_builder.parse(m_database);
            database_document.renameNode(database_document.getDocumentElement(), "",
                                         "FilamentSensorDatabase");
            Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            FileWriter writer = new FileWriter(m_database);
            transformer.transform(new DOMSource(database_document), new StreamResult(writer));
            writer.close();
            return true;
        }
        catch (Exception e) // Catch generic Exception to avoid redundant code
        { // Can be: IO-, SAX- (parse error) or IllegalArgumentException
            System.err.println("Could not read '" + m_database + "':\n" + e.getMessage());
        }
        return false;
    }

    /**
     * Read the config file and parse the XML
     *
     * @return true if successful, false otherwise
     */
    private static boolean readConfig(String config_file)
    {
        try
        {
            m_config_document = m_document_builder.parse(config_file);
            Element root = m_config_document.getDocumentElement();
            m_config_document.renameNode(root, "", "FilamentSensor");
        }
        catch (Exception e) // Catch generic Exception to avoid redundant code
        { // Can be: IO-, SAX- (parse error) or IllegalArgumentException
            if (FilamentSensor.DEBUG)
            {
                System.out.println("Could not read '" + config_file + "':\n" + e.getMessage());
            }
            return false;
        }
        if (FilamentSensor.DEBUG)
        {
            System.out.println("Config file read and successfully parsed.");
        }

        return true;
    }

    /**
     * Save the configuration to file
     *
     * @return true if successful, false otherwise
     */
    public static boolean saveConfig()
    {
        if (m_configuration_file_path.equals(""))
        {
            return false;
        }

        try
        {
            File configuration_file = new File(m_configuration_file_path);

            // create the file if it doesn't already exist
            if (configuration_file.createNewFile() && FilamentSensor.DEBUG)
            {
                System.out.println("Config file created.");
            }

            // set up the Transformer object
            Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");

            FileWriter writer = new FileWriter(configuration_file);
            StreamResult result = new StreamResult(writer);
            DOMSource source = new DOMSource(m_config_document);
            transformer.transform(source, result);
            writer.close();
        }
        catch (Exception e)
        {
            System.out.println("Saving failed: " + e.getMessage());
            if (FilamentSensor.DEBUG)
            {
                System.out.println("test: " + e.toString());
            }
            return false;
        }
        if (FilamentSensor.DEBUG)
        {
            System.out.println("Configuration successfully saved.");
        }

        return true;
    }

    /**
     * Reads paths of files of the given type from the database and returns
     * them as File objects. Does not check for existence of files on disk.
     */
    public static List<File> readDatabase(int type)
    {
        Document database_document = null;
        try
        {
            database_document = m_document_builder.parse(m_database);
        }
        catch (Exception e) // Catch generic Exception to avoid redundant code
        { // Can be: IO-, SAX- (parse error) or IllegalArgumentException
            System.out.println("Could not read '" + m_database + "':\n" + e.getMessage());
            return null;
        }

        NodeList element_list = database_document.getElementsByTagName("FilamentSensorDatabase");
        if (element_list.getLength() == 0)
        {
            element_list = database_document.getElementsByTagName("FilTrackerDatabase");
            if (element_list.getLength() == 0)
            {
                System.out.println("The specified file is valid XML, but does not seem to be a valid "
                                   + "FilamentSensor database...");
                return null;
            }
        }
        
        String file_type = m_type_map.get(type);
        if (file_type == null)
        {
            return null;
        }
        
        element_list = database_document.getElementsByTagName(file_type);
        if (element_list == null)
        {
            return null;
        }

        List<File> list = new ArrayList<File>();

        if (element_list.getLength() != 0)
        {
            for (Node n = element_list.item(0).getFirstChild(); n != null; n = n.getNextSibling())
            {
                try
                {
                    if (n.getNodeType() != Node.ELEMENT_NODE)
                    {
                        continue;
                    }

                    Element element = (Element) n;
                    list.add(new File(element.getAttribute("file")));
                }
                catch (Exception e)
                {
                    if (FilamentSensor.DEBUG)
                    {
                        System.out.println("Error parsing file name '" + n + "': " + e.getMessage());
                    }
                }
            }
        }
        
        return list;
    }

    /**
     * Adds the given path to the database for the given type if it is not
     * yet present. Returns true if successful, false otherwise.
     */
    public static boolean saveToDatabase(String path, int type)
    {
        List<File> list = readDatabase(type);
        if (list != null && list.contains(new File(path)))
        {
            return true;
        }
        
        Document database_document = null;
        try
        {
            database_document = m_document_builder.parse(m_database);
            database_document.renameNode(database_document.getDocumentElement(), "",
                                         "FilamentSensorDatabase");
        }
        catch (Exception e) // Catch generic Exception to avoid redundant code
        { // Can be: IO-, SAX- (parse error) or IllegalArgumentException
            System.out.println("Could not read '" + m_database + "':\n" + e.getMessage());
            return false;
        }
        
        String file_type = m_type_map.get(type);
        if (file_type == null)
        {
            System.out.println("Invalid file type: " + file_type);
            return false;
        }

        NodeList n_list = database_document.getDocumentElement().getElementsByTagName(file_type);
        if (n_list.getLength() > 1)
        {
            System.out.println("More than one element with tag " + file_type + ". Cannot process.");
            return false;
        }
        if (n_list.getLength() == 0)
        {
            database_document.getDocumentElement().appendChild(database_document.createElement(file_type));
            n_list = database_document.getDocumentElement().getElementsByTagName(file_type);
        }

        Node type_element = n_list.item(0);
        int n = countChildren(type_element);
        Element file_element = database_document.createElement("file" + n);
        type_element.appendChild(file_element);
        file_element.setAttribute("file", path);

        try
        {
            // write the created document object to file
            Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");

            FileWriter writer = new FileWriter(m_database);
            transformer.transform(new DOMSource(database_document), new StreamResult(writer));
            writer.close();
        }
        catch (Exception e) // IOException or TransformerConfigurationException
        {
            System.out.println("Error while writing '" + m_database + "': " + e.getMessage());
            return false;
        }
        return true;
    }

    public static int countChildren(Node node)
    {
        int counter = 0;
        for (Node child = node.getFirstChild(); child != null; child = child.getNextSibling())
        {
            if (child.getNodeType() == Node.ELEMENT_NODE || child.hasAttributes())
            {
                counter++;
            }
        }
        return counter;
    }
    
    private static Element findElement(String element, boolean create)
    {
        Element el;
        NodeList list = m_config_document.getElementsByTagName(element);

        if (list.getLength() == 0)
        {
            if (create)
            {
                el = m_config_document.createElement(element);
                m_config_document.getDocumentElement().appendChild(el);
            }
            else
            {
                return null;
            }
        }
        else
        {
            el = (Element) list.item(0);
        }

        if (list.getLength() > 1 && FilamentSensor.DEBUG)
        {
            System.out.println("Warning: multiple Elements with tagName '" + element
                               + "' found. Only the first one is used.");
        }

        return el;
    }

    public static String get(String element, String attribute)
    {
        Element el = findElement(element, false);
        String value = "";

        if (el != null)
        {
            value = el.getAttribute(attribute);
        }

        if (value.equals(""))
        {
            value = defaults.get(element + "." + attribute);
        }

        if (value == null)
        {
            value = "";
        }

        return value;
    }

    public static int getInt(String element, String attribute)
    {
        int value = 0;
        try
        {
            value = Integer.parseInt(get(element, attribute));
        }
        catch (NumberFormatException e)
        {
            if (FilamentSensor.DEBUG)
            {
                System.out.println("Error: '" + element + "." + attribute
                                   + "' does not appear to be an integer-type! Returning 0...");
            }
        }
        return value;
    }

    public static void set(String element, String attribute, String value)
    {
        Element el = findElement(element, true);
        el.setAttribute(attribute, value);
    }

    public static void set(String element, String attribute, int value)
    {
        set(element, attribute, "" + value);
    }

}
