package ui;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;
import javax.swing.JFileChooser;


public abstract class Dialogs
{
    /**
     * Generic dialog for opening a single file or a number of files from
     * a folder.
     *
     * @param folder    The dialog returns a folder if true a file otherwise.
     * @param path      Starting Folder shown in the dialog.
     * @param caption   Window caption of the dialog.
     * @return          A File object containing the path of the selected
     *                  file or folder.
     */
    public static File load(boolean folder, String path, String caption)
    {
        return IODialog(folder, path, caption, false, "Open");
    }

    public static File load(boolean folder, File path, String caption)
    {
        return load(folder, path.getAbsolutePath(), caption);
    }

    public static File load(boolean folder, String caption)
    {
        return load(folder, Config.get("LastFolders", "Open"), caption);
    }

    /**
     * Generic dialog for saving a single file or a number of files to
     * a folder.
     *
     * @param folder    The dialog returns a folder if true a file otherwise.
     * @param path      Starting Folder shown in the dialog.
     * @param caption   Window caption of the dialog.
     * @return          A File object containing the path of the selected
     *                  file or folder.
     */
    public static File save(boolean folder, String path, String caption)
    {
        return IODialog(folder, path, caption, true, "Save");
    }

    public static File save(boolean folder, String caption)
    {
        return save(folder, Config.get("LastFolders", "Save"), caption);
    }

    public static boolean saveImage(BufferedImage image)
    {
        File image_file = save(false, "Save image...");
        if (image_file == null)
        {
            return false;
        }
        try
        {
            if (FilamentSensor.DEBUG)
            {
                System.out.println("Try to save image to '" + image_file.getAbsolutePath() + "'.");
            }
            return ImageIO.write(image, "png", image_file);
        }
        catch (IOException e)
        {
            System.out.println(e.getMessage());
            return false;
        }
    }

    private static File IODialog(boolean folder, String path, String caption, boolean save, String name)
    {
        JFileChooser fc = new JFileChooser();
        fc.setDialogTitle(caption);
        fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
        fc.setAcceptAllFileFilterUsed(false);

        if (path == "")
        {
            // work around bug 7130662
            path = System.getProperty("user.home");
        }

        File pathFile = new File(path);
        fc.setCurrentDirectory(pathFile);

        if (folder)
        {
            fc.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
            if ((save ? fc.showSaveDialog(MainWindow.frame()) :
                        fc.showOpenDialog(MainWindow.frame())) == JFileChooser.APPROVE_OPTION)
            {
                File file = fc.getSelectedFile();

                if (file == null)
                {
                    return null;
                }

                if (file.isDirectory())
                {
                    Config.set("LastFolders", name, fc.getSelectedFile().getAbsolutePath());
                    return fc.getSelectedFile();
                }
                Config.set("LastFolders", name, fc.getSelectedFile().getParent());
                return fc.getSelectedFile().getParentFile();
            }
            return null;
        }

        if ((save ? fc.showSaveDialog(MainWindow.frame()) :
                    fc.showOpenDialog(MainWindow.frame())) == JFileChooser.APPROVE_OPTION)
        {
            if (fc.getSelectedFile() == null)
            {
                return null;
            }

            Config.set("LastFolders", name, fc.getSelectedFile().getParent());
            return fc.getSelectedFile();
        }
        return null;
    }
}
