package ui;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.awt.Point;

import javax.swing.JOptionPane;
import javax.swing.JFileChooser;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;

import core.CellData;
import core.Const;
import core.DataFilaments;
import core.DataImage;
import core.DataParameters;
import core.DataParameters.Any;
import core.DataProject;
import core.Filament;
import core.FilamentChain;
import core.Misc;

public class IOImageData implements Misc
{
    private File m_xml_file;

    private Document m_xml_source_document;
    private DocumentBuilder m_document_builder;
    private Element m_xml_root_element;

    private static String m_path_separator;

    /**
     * Create an object to load or save image data XML files. There is no check
     * whatsoever, that the passed file object points to a valid data file. If
     * it does not, read() will fail.
     */
    public IOImageData(File file)
    {
        m_xml_file = file;
        m_path_separator = System.getProperty("file.separator");

        try
        {
            // Initialize documentBuilder
            DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
            m_document_builder = docFactory.newDocumentBuilder();
        }
        catch (ParserConfigurationException e)
        {
            System.out.println("Fatal: failed to create DocumentBuilder! " + e.getMessage());
            m_document_builder = null;
        }

        if (FilamentSensor.DEBUG)
        {
            System.out.println("IOImageData object created ('" + m_xml_file + "').");
        }
    }

    String getName()
    {
        if (m_xml_file == null)
        {
            return null;
        }

        return m_xml_file.getName();
    }

    String getPath()
    {
        return m_xml_file.getPath();
    }

    String getParentFolder()
    {
        return m_xml_file.getParent();
    }

    void setTarget(File file)
    {
        m_xml_file = file;
    }

    public boolean writeImageDataFile(DataImage data)
    {
        if (m_document_builder == null)
        {
            System.out.println("Fatal: Can't save, docBuilder == null!");
            return false;
        }

        m_xml_source_document = m_document_builder.newDocument();
        m_xml_root_element = m_xml_source_document.createElement("FilamentSensorImageDataFile");
        m_xml_source_document.appendChild(m_xml_root_element);

        // Image file data
        Element element = m_xml_source_document.createElement("img_file");
        m_xml_root_element.appendChild(element);
        element.setAttribute("name", data.getImageName());
        element.setAttribute("path", getRelativePath(m_xml_file, data.getFile()));
        element.setAttribute("resolution", data.getResolution());
        element.setAttribute("dpi", "not_implemented"); // TODO

        // Cell data
        element = m_xml_source_document.createElement("cell_data");
        m_xml_root_element.appendChild(element);
        element.setAttribute("center_x", "" + data.getCellData().getCenter().x);
        element.setAttribute("center_y", "" + data.getCellData().getCenter().y);
        element.setAttribute("area", "" + data.getCellData().getArea());
        element.setAttribute("aspect_ratio", "" + data.getCellData().getAspectRatio());
        element.setAttribute("long_half_axis", "" + data.getCellData().getLongHalfAxis());
        element.setAttribute("orientation", "" + data.getCellData().getOrientation());
        element.setAttribute("order_parameter", "" + data.getCellData().getOrderParameter());

        if (!data.usesDefaultParameters())
        {
            for (int type : DataParameters.ALL)
            {
                element = m_xml_source_document.createElement(DataParameters.getName(type));
                m_xml_root_element.appendChild(element);
                for (Any key : data.keySet(type))
                {
                    element.setAttribute(key.name(), "" + data.getParameter(key));
                }
            }
        }

        Element chains = m_xml_source_document.createElement("filament_chains");
        m_xml_root_element.appendChild(chains);

        for (int i = 0; i < data.countChains(); i++)
        {
            chains.appendChild(data.getFilamentChain(i).toXML(m_xml_source_document, i, 0));
        }

        Element fields = m_xml_source_document.createElement("orientation_fields");
        m_xml_root_element.appendChild(fields);

        try
        {
            // write the created document object to file
            Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");

            FileWriter writer = new FileWriter(m_xml_file);
            StreamResult result = new StreamResult(writer);
            DOMSource source = new DOMSource(m_xml_source_document);
            transformer.transform(source, result);
            writer.close();
        }
        catch (Exception e) // IOException or
                            // TransformerConfigurationException
        {
            System.out.println("Error while saving '" + m_xml_file + "': " + e.getMessage());
            return false;
        }

        return Config.saveToDatabase(m_xml_file.getAbsolutePath(), Config.IMAGE);
    }

    File getImage()
    {
        if (m_document_builder == null)
        {
            return null;
        }

        try
        {
            m_xml_source_document = m_document_builder.parse(m_xml_file);
            m_xml_root_element = m_xml_source_document.getDocumentElement();
        }
        catch (Exception e) // Catch generic Exception to avoid redundant code
        { // Can be: IO-, SAX- (parse error) or IllegalArgumentException
            System.out.println("Could not read '" + m_xml_file + "':\n" + e.getMessage());
            return null;
        }

        NodeList fplist = m_xml_source_document.getElementsByTagName("FilPickerProject");
        NodeList ftlist = m_xml_source_document.getElementsByTagName("FilTrackerImageDataFile");
        NodeList fslist = m_xml_source_document.getElementsByTagName("FilamentSensorImageDataFile");

        if (fplist.getLength() != 0 || ftlist.getLength() != 0 || fslist.getLength() != 0)
        {
            File image_file;
            Element element;

            // Image file data
            NodeList list = m_xml_source_document.getElementsByTagName("img_file");
            if (list.getLength() == 0)
            {
                image_file = null;
            }
            else
            {
                element = (Element) list.item(0);
                String path = element.getAttribute("path"), name = element.getAttribute("name");

                image_file = new File(path);

                if (!image_file.isAbsolute())
                {
                    path = m_xml_file.getParent() + m_path_separator + path + m_path_separator + name;
                }
                else
                {
                    // support old project files
                    if (name != "")
                    {
                        path += m_path_separator + name;
                    }
                }

                image_file = new File(path);

                try
                {
                    image_file = image_file.getCanonicalFile();
                }
                catch (IOException e)
                {
                    if (FilamentSensor.DEBUG)
                    {
                        System.out.println("Converting '" + path + "' to canonical path failed:\n" + e.getMessage());
                    }
                }
            }
            return image_file;
        }

        System.out.println("The specified file " + m_xml_file + " is valid XML, but does not seem to be a valid FilamentSensor-File...");
        return null;
    }

    /**
     * This method reads a data XML file for a single image and returns an
     * object representing the data.
     *
     * @param default_params
     *            Default image processing parameters. These will be used, if
     *            none are given in the file or if force_defaults is true.
     * @param force_defaults
     *            If true, the given default parameters will be used, even if
     *            the XML file contains parameters.
     * @return An object containing a File object representing the image file, a
     *         set of parameters and possibly filament chain data.
     */
    public DataImage read(String image_path, DataParameters default_params, boolean force_defaults)
    {
        if (m_document_builder == null)
        {
            return null;
        }

        try
        {
            m_xml_source_document = m_document_builder.parse(m_xml_file);
            m_xml_root_element = m_xml_source_document.getDocumentElement();
        }
        catch (Exception e) // Catch generic Exception to avoid redundant code
        { // Can be: IO-, SAX- (parse error) or IllegalArgumentException
            System.out.println("Could not read '" + m_xml_file + "':\n" + e.getMessage());
            return null;
        }

        NodeList list = m_xml_source_document.getElementsByTagName("FilPickerProject");
        if (list.getLength() != 0)
        {
            return readFilPicker(default_params, force_defaults);
        }

        NodeList list_old = m_xml_source_document.getElementsByTagName("FilTrackerImageDataFile");
        list = m_xml_source_document.getElementsByTagName("FilamentSensorImageDataFile");
        if (list_old.getLength() != 0 || list.getLength() != 0)
        {
            return readDataFile(image_path, default_params, force_defaults);
        }

        System.err.println("The specified file " + m_xml_file.getAbsolutePath() + " is valid XML, but does not seem to be a valid FilamentSensor-File...");
        return null;
    }

    private DataImage readFilPicker(DataParameters default_params, boolean force_defaults)
    {
        File image_file = readImageFile();
        String message = null;

        // If no image file is given in the xml file or the file was not found.
        if (image_file == null)
        {
            message = "This project does not specifiy an image file.\n" +
                      "Would you like to import one?";
        }
        else if (!image_file.exists())
        {
            message = "The file \"" + image_file + "\" was not found.\n" +
                      "Would you like to search for it?";
        }
        else if (!image_file.canRead())
        {
            message = "The file \"" + image_file + "\" could not be read.\n" +
                       "Would you like to import another image?";
        }

        if (message != null)
        {
            if (JOptionPane.showOptionDialog(MainWindow.frame(), message, "Browse for image file?",
                    JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE, null, null, null) != JOptionPane.YES_OPTION)
            {
                return null;
            }

            ImageFileChooser ifc = new ImageFileChooser();

            String folder = Config.get("LastFolders", "Import");
            if (folder != "")
            {
                ifc.setCurrentDirectory(new File(folder));
            }

            if (ifc.showOpenDialog(MainWindow.frame()) != JFileChooser.APPROVE_OPTION)
            {
                return null;
            }

            image_file = ifc.getSelectedFile();
        }

        //////////////// XML read, now parse it into FPData ///////////////////////////

        DataParameters parameters = (force_defaults ? default_params :
                                     DataParameters.fromXML(m_xml_source_document, default_params));
        parameters.completeFrom(DataProject.getDefaultParameters()).completeFrom(new DataParameters());

        // Filament list.
        NodeList list = m_xml_source_document.getElementsByTagName("filaments");
        List<Filament> filament_list = new ArrayList<Filament>();
        for (Node n = list.item(0).getFirstChild(); n != null; n = n.getNextSibling())
        {
            try
            {
                if (n.getNodeType() != Node.ELEMENT_NODE)
                {
                    continue;
                }

                Element element = (Element) n;
                Filament fil = new Filament(new Point(Integer.parseInt(element.getAttribute("start_x")),
                                                      Integer.parseInt(element.getAttribute("start_y"))),
                                            new Point(Integer.parseInt(element.getAttribute("end_x")),
                                                      Integer.parseInt(element.getAttribute("end_y"))),
                                            Math.round(Const.M * Double.parseDouble(element.getAttribute("width"))),
                                            element.getAttribute("keep").equals("yes"));
                filament_list.add(fil);
            }
            catch (Exception e)
            {
                if (FilamentSensor.DEBUG)
                {
                    System.out.println("Error parsing Filament '" + n + "': " + e.getMessage());
                }
            }
        }
        DataImage data = new DataImage(image_file, null, parameters,
                                       new DataFilaments(filament_list));

        if (FilamentSensor.DEBUG)
        {
            System.out.println("XML successfully parsed, imported " + data.countFilaments() +
                               " filaments.");
        }

        return data;
    }

    private DataImage readDataFile(String image_path, DataParameters default_params, boolean force_defaults)
    {
        // TODO: Implement a switch that allows to skip reading the image file?
        if(image_path == null)
        {
            image_path = "";
        }
        File image_file = readImageFile();
        String message = null;

        // If no image file is given in the xml file or the file was not found.
        if (image_file == null)
        {
            message = "This data file does not specifiy an image file.\n" +
                      "Would you like to import one?";
        }
        else if (!image_file.exists())
        {
            File tmp = new File(image_path + m_path_separator + image_file.getName());
            if (tmp.exists() && tmp.canRead())
            {
                image_file = tmp;
            }
            else
            {
                message = "The file \"" + image_file + "\" was not found.\n" +
                          "Would you like to search for it?";
            }
        }
        else if (!image_file.canRead())
        {
            File tmp = new File(image_path + m_path_separator + image_file.getName());
            if (tmp.exists() && tmp.canRead())
            {
                image_file = tmp;
            }
            else
            {
                message = "The file \"" + image_file + "\" could not be read.\n" +
                          "Would you like to import another image?";
            }
        }

        if (message != null)
        {
            if (JOptionPane.showOptionDialog(MainWindow.frame(), message, "Browse for image file?",
                    JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE, null, null, null) != JOptionPane.YES_OPTION)
            {
                return null;
            }

            ImageFileChooser ifc = new ImageFileChooser();

            String folder = Config.get("LastFolders", "Import");
            if (folder != "")
            {
                ifc.setCurrentDirectory(new File(folder));
            }

            if (ifc.showOpenDialog(MainWindow.frame()) != JFileChooser.APPROVE_OPTION)
            {
                return null;
            }

            image_file = ifc.getSelectedFile();
        }

        NodeList list = m_xml_source_document.getElementsByTagName("cell_data");
        CellData cell_data = null;
        if (list.getLength() > 0)
        {
            cell_data = new CellData((Element) list.item(0));
        }

        DataParameters parameters = (force_defaults ? default_params :
                                     DataParameters.fromXML(m_xml_source_document, default_params));
        parameters.completeFrom(DataProject.getDefaultParameters()).completeFrom(new DataParameters());

        // Filament data
        list = m_xml_source_document.getElementsByTagName("filament_chains");
        List<FilamentChain> chain_list = new ArrayList<FilamentChain>();

        // Iterate through filament chains.
        for (Node chain = list.item(0).getFirstChild(); chain != null; chain = chain.getNextSibling())
        {
            if (!chain.hasAttributes() || Long.parseLong(((Element) chain).getAttribute("length")) == 0)
            {
                continue;
            }
            chain_list.add(new FilamentChain(chain));
        }
        DataFilaments filaments = new DataFilaments(chain_list, true);

        // TODO: Account for the case of no import of image data?
        DataImage data = new DataImage(image_file, cell_data, parameters, filaments);

        if (FilamentSensor.DEBUG)
        {
            System.out.println("XML successfully parsed, imported " + data.countChains() + " filament chains.");
        }

        return data;
    }

    private File readImageFile()
    {
        NodeList list = m_xml_source_document.getElementsByTagName("img_file");
        if (list.getLength() == 0)
        {
            return null;
        }

        Element element = (Element) list.item(0);
        String path = element.getAttribute("path"), name = element.getAttribute("name");
        File image_file = new File(path);

        if (image_file.isAbsolute())
        {
            // support old project files
            if (name != "")
            {
                path += m_path_separator + name;
            }
        }
        else
        {
            path = m_xml_file.getParent() + m_path_separator + path + m_path_separator + name;
        }

        image_file = new File(path);

        try
        {
            image_file = image_file.getCanonicalFile();
        }
        catch (IOException e)
        {
            if (FilamentSensor.DEBUG)
            {
                System.out.println("Converting '" + path + "' to canonical path failed:\n" + e.getMessage());
            }
        }
        return image_file;
    }

    /**
     * Gives the path of the method's argument relative to the path of
     * m_xml_file.
     */
    private static String getRelativePath(File origin, File target)
    {
        if (origin == null || target == null)
        {
            return null;
        }

        File origin_ancestor, target_ancestor;
        String path_bak = "", path_fwd = "";

        for (origin_ancestor = origin.getParentFile(); origin_ancestor != null; origin_ancestor = origin_ancestor.getParentFile())
        {
            for (target_ancestor = target.getParentFile(); target_ancestor != null; target_ancestor = target_ancestor.getParentFile())
            {
                if (origin_ancestor.equals(target_ancestor))
                {
                    // don't return relative path if the common parent is the fs-root
                    if (origin_ancestor.getParentFile() == null)
                    {
                        break;
                    }

                    if (path_bak + path_fwd == "")
                    {
                        return ".";
                    }

                    return path_bak + path_fwd;
                }
                path_fwd = target_ancestor.getName() + (path_fwd.equals("") ? "" : m_path_separator + path_fwd);
            }
            path_fwd = "";
            path_bak += ".." + m_path_separator;
        }

        return target.getParent();
    }
}
