package ui;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferInt;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;

import core.Misc.Int2D;
import core.Misc.Pair;

public abstract class ImageFactory
{

    /**
     * Converts the array of pixels to BufferedImage.
     */
    public static BufferedImage makeBinaryImage(boolean[][] binary_image, boolean invert)
    {
        if (binary_image == null || binary_image.length <= 0 || binary_image[0].length <= 0)
        {
            return null;
        }

        int width = binary_image.length;
        int height = binary_image[0].length;
        BufferedImage image = new BufferedImage(width, height, BufferedImage.TYPE_BYTE_GRAY);
        WritableRaster raster = image.getRaster();

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                raster.setSample(i, j, 0, (binary_image[i][j] == invert ? 0 : 255));
            }
        }

        return image;
    }
    
    public static BufferedImage makeGreyImage(int[][] int_image, boolean filter)
    {
        if (int_image == null || int_image.length <= 0 || int_image[0].length <= 0)
        {
            return null;
        }

        int width = int_image.length;
        int height = int_image[0].length;
        BufferedImage image = new BufferedImage(width, height, BufferedImage.TYPE_BYTE_GRAY);
        WritableRaster raster = image.getRaster();

        if (filter)
        {
            for (int i = 0; i < width; i++)
            {
                for (int j = 0; j < height; j++)
                {
                    int pixel = int_image[i][j] + 128;
                    raster.setSample(i, j, 0, (pixel < 0 ? 0 :
                                              (pixel > 255 ? 255 :
                                                  pixel)));
                }
            }
            return image;
        }

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                raster.setSample(i, j, 0, (int_image[i][j] < 0 ? 0 :
                                          (int_image[i][j] > 255 ? 255 :
                                               int_image[i][j])));
            }
        }
        return image;
    }

    public static BufferedImage makeColorImage(int[][] int_image, int color_count)
    {
        if (int_image == null || int_image.length <= 0 || int_image[0].length <= 0)
        {
            return null;
        }

        final int width = int_image.length,
                  height = int_image[0].length;
        int[] out = new int[width * height],
              colors = makeColors(color_count);

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                out[width * j + i] = (int_image[i][j] < 0 ? 0xFF000000 :
                                     (int_image[i][j] >= color_count ? 0xFFFFFFFF :
                                         colors[int_image[i][j]]));
            }
        }

        ColorModel cm = ColorModel.getRGBdefault();
        DataBuffer buffer = new DataBufferInt(out, width * height);
        WritableRaster raster = Raster.createWritableRaster(cm.createCompatibleSampleModel(width, height), buffer, null);
        return new BufferedImage(cm, raster, false, null);
    }

    public static BufferedImage makeWidthMap(Pair<Int2D, Integer> pair)
    {
        int[][] width_map = pair.a.the;
        final int color_count = pair.b - 1,
                  width = width_map.length,
                  height = width_map[0].length,
                  offset = 5 * (color_count + 1) / 3 + 1;
        int[] out = new int[width * height],
              colors = makeColors(color_count);

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                out[width * j + i] = (width_map[i][j] == 0 ? 0xFF000000 :
                                     (width_map[i][j] == 1 ||
                                      width_map[i][j] > color_count + 1 ? 0xFFFFFFFF :
                                         colors[((offset - width_map[i][j]) % color_count)]));
            }
        }

        ColorModel cm = ColorModel.getRGBdefault();
        DataBuffer buffer = new DataBufferInt(out, width * height);
        WritableRaster raster = Raster.createWritableRaster(cm.createCompatibleSampleModel(width, height), buffer, null);
        return new BufferedImage(cm, raster, false, null);
    }

    public static BufferedImage makeColorImage(int[][] int_image)
    {
        return makeColorImage(int_image, 180);
    }

    public static int[] makeColors(int count)
    {
        int current;
        int[] rainbow = new int[count];
        double step = 1530.0 / count;

        for (int i = 0; i < count; i++)
        {
            current = (int) (i * step);

            int r = (current < 511 ? 255 :
                    (current < 766 ? 765 - current :
                    (current < 1021 ? 0 :
                    (current < 1276 ? current - 1020 :
                        255))));
            int g = (current < 511 ? current/2 :
                    (current < 766 ? 255 :
                    (current < 1021 ? 1020 - current :
                        0)));
            int b = (current < 766 ? 0 :
                    (current < 1021 ? current - 765 :
                    (current < 1276 ? 255 :
                        1530 - current)));

            rainbow[i] = 0xFF000000 | (r << 16) | (g << 8) | b;
        }

        return rainbow;
    }

}
