package ui;
/*
 * An extension of JFileChooser to browse for image files with a preview thumbnail.
 *
 * Copyright 2012 Julian Rüger
 *
 * This class is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This class is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this class; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.io.File;
import java.io.IOException;

import java.awt.Dimension;
import java.awt.Graphics;

import java.awt.image.BufferedImage;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.JComponent;
import javax.swing.JFileChooser;

import javax.swing.filechooser.FileFilter;

import javax.imageio.ImageIO;


public class ImageFileChooser extends JFileChooser
{
    public ImageFileChooser()
    {
        super();
        ImageFileFilter filter = new ImageFileFilter();
        setAccessory(new ImagePreview(this));
        addChoosableFileFilter(filter);
        setFileFilter(filter);
    }

    public static BufferedImage readImage(File file)
    {
        if (file == null)
        {
            return null;
        }

        BufferedImage img = null;

        try
        {
            img = ImageIO.read(file);
        }
        catch (IOException e)
        {
            if (FilamentSensor.DEBUG)
            {
                System.out.println("ImageIO.read failed: " + e + "\nTrying fallback method.");
            }
        }

        /*
         *  If reading image file failed, try to use ImageMagick
         *  to create a temporary png file.
         */
        if (img == null || img.getWidth() < 0 || img.getHeight() < 0)
        {
            img = null;
            File tmp_file = null;

            try
            {
                Process p;
                tmp_file = File.createTempFile("filpicker", ".png");

                p = Runtime.getRuntime().exec("convert " + file.getPath() + " " + tmp_file.getPath());
                p.waitFor();

                img = ImageIO.read(tmp_file);
            }
            catch (Exception e)
            {
                if (FilamentSensor.DEBUG)
                {
                    System.out.println("Fallback failed: " + e);
                }
            }

            if (img != null && (img.getWidth() < 0 || img.getHeight() < 0))
            {
                img = null;
            }

            try
            {
                tmp_file.deleteOnExit();
            }
            catch (Exception e)
            {
                // SecurityException, NullPointerException, IOException
            }
        }

        return img;
    }

    private class ImageFileFilter extends FileFilter
    {
        @Override
        public boolean accept(File file)
        {
            if (file.isDirectory())
            {
                return true;
            }

            String sfile = file.getName().toLowerCase();
            int i = sfile.lastIndexOf('.');
            String ext = i < 0 ? "" : sfile.substring(i + 1, sfile.length());

            return ext.equals("png") || ext.equals("pgm") || ext.equals("tif") || ext.equals("tiff")
                    || ext.equals("bmp") || ext.equals("jpg") || ext.equals("jpeg") || ext.equals("gif")
                    || ext.equals("eps");
        }

        @Override
        public String getDescription()
        {
            return "Image files";
        }
    }

    private class ImagePreview extends JComponent implements PropertyChangeListener
    {
        private static final int PREFERRED_WIDTH = 200, PREFERRED_HEIGHT = 300;

        private BufferedImage thumb;
        private File file;

        private ImagePreview(JFileChooser fc)
        {
            super();
            setPreferredSize(new Dimension(PREFERRED_WIDTH + 10, PREFERRED_HEIGHT));
            fc.addPropertyChangeListener(this);
        }

        @Override
        public void propertyChange(PropertyChangeEvent e)
        {
            boolean update = false;
            String prop = e.getPropertyName();

            // If the directory changed, don't show an image
            if (JFileChooser.DIRECTORY_CHANGED_PROPERTY.equals(prop))
            {
                file = null;
                update = true;
            }
            else if (JFileChooser.SELECTED_FILE_CHANGED_PROPERTY.equals(prop))
            {
                file = (File) e.getNewValue();
                update = true;
            }

            // Update the preview
            if (update)
            {
                thumb = null;
                if (isShowing())
                {
                    thumb = readImage(file);
                    this.repaint();
                }
            }
        }

        @Override
        protected void paintComponent(Graphics g)
        {
            if (thumb != null)
            {
                int w = thumb.getWidth(), h = thumb.getHeight(), w2 = Math.min(w, PREFERRED_WIDTH), h2 = Math.min(h,
                        PREFERRED_HEIGHT);

                double factor = Math.min(w2 / (double) w, h2 / (double) h);

                w2 = (int) Math.round(factor * w);
                h2 = (int) Math.round(factor * h);

                int x = getWidth() / 2 - w2 / 2, y = getHeight() / 2 - h2 / 2;

                if (y < 0)
                {
                    y = 0;
                }
                if (x < 10)
                {
                    x = 10;
                }

                g.drawImage(thumb, x, y, x + w2, y + h2, 0, 0, w, h, this);
            }
        }
    }
}
