package ui;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2014 Julian Rüger
 *               2013      Patricia Burger, Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.BoxLayout;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.WindowConstants;

import core.DataProject;

/**
 * Main class singleton. Contains the main window (JFrame).
 */
public class MainWindow extends WindowAdapter implements ActionListener
{
    static final int MIN_WIDTH = 560, MIN_HEIGHT = 340;
    static final String LOOKANDFEEL = lookAndFeel();
    private static final String INFO = 
        "FilamentSensor has no documentation yet. Most things should be self-explanatory.\n" +
        "However, please note the following:\n\n" +
        "   * FilamentSensor considers folders as projects and opens all images in a folder.\n\n" +
        "   * FilamentSensor's interface is unresponsive while the program calculates.\n" +
        "     Opening a project or a new folder will take several seconds.\n" +
        "     Batch processing takes about 20 seconds per image so the interface will\n" +
        "     be unresponsive for some time during batch processing.\n\n" +
        "   * FilamentSensor supports the png, jpg, bmp and gif formats. It does NOT support tif!\n\n" +
        "   * FilamentSensor expects grayscale images and will convert color images to grayscale.\n\n" +
        "Basic functionality:\n\n" +
        "   * Start a new project via the menu: File -> New images... and choose a folder.\n" +
        "     Selecting a single file will import all images in its folder.\n\n" +
        "   * To process an image, double click on the thumbnail.\n\n" +
        "   * Note that changing parameters for image processing will automatically trigger\n" +
        "     reprocessing, which takes some time. Using the spinboxes is thus much slower\n" +
        "     than typing in the desired value.\n\n" +
        "   * The eLoG tab does not display an image unless eLoG processing has been done.\n\n" +
        "   * This information is available via the help menu.\n\n";

    // Single instance
    private static MainWindow the;

    // main window controls
    private final JFrame frame = new JFrame(FilamentSensor.TITLE + " " + FilamentSensor.VERSION);
    private TabBar tabBar;
    private JPanel main;

    /**
     * Constructor of the main class. It invokes the subclasses' constructors,
     * initializes and paints the main window.
     */
    private MainWindow()
    {
        super();

        Config.init(".filamentsensor");

        JFrame.setDefaultLookAndFeelDecorated(false);
        try
        {
            UIManager.setLookAndFeel(LOOKANDFEEL);
            SwingUtilities.updateComponentTreeUI(frame);
            frame.pack();
        }
        catch (Exception e)
        /*
         *  Catch generic Exception to avoid redundant code
         *  Can be: ClassNotFound-, Instantiation-, IllegalAccess- or
         *  UnsupportedLookAndFeelException
         */
        {
            System.out.println("Could not set look and feel:\n" + e.getMessage());
        }

        // Creates main menu with "File", "Edit", "View" ...
        MainWindowMenu.create(this);
        frame.setJMenuBar(MainWindowMenu.get());

        frame.setContentPane(createContentPane());
        frame.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
        frame.addWindowListener(this);

        frame.setMinimumSize(new Dimension(MIN_WIDTH, MIN_HEIGHT));
        frame.setSize(Config.getInt("MainWnd", "size_w"), Config.getInt("MainWnd", "size_h"));
        frame.setLocation(Config.getInt("MainWnd", "pos_w"), Config.getInt("MainWnd", "pos_h"));
        frame.setVisible(true);

        if (Config.getInt("FilamentSensor", "show_intro") == 1)
        {
            intro();
        }
    }

    static void open()
    {
        if (the == null)
        {
            the = new MainWindow();
        }
    }

    public static JFrame frame()
    {
        if (the == null)
        {
            return null;
        }
        return the.frame;
    }

    /**
     * Create main window content elements, set text, size, ActionHandler, etc.
     *
     * @return JPanel containing the main window elements.
     */
    private JPanel createContentPane()
    {
        main = new JPanel();
        main.setLayout(new BoxLayout(main, BoxLayout.Y_AXIS));

        tabBar = new TabBar(MainWindowMenu.get());
        TabBar.tab_divider_location = Config.getInt("tabBar", "divider_location");
        main.add(tabBar);

        if (FilamentSensor.DEBUG)
        {
            System.out.println("Main window content created.");
        }

        return main;
    }
    
    private static void intro()
    {
        JCheckBox checkbox = new JCheckBox("Do not show this message again.");
        Object[] params = {INFO, checkbox};
        JOptionPane.showMessageDialog(MainWindow.frame(), params, "Welcome to FilamentSensor",
                                      JOptionPane.INFORMATION_MESSAGE, null);
        if (checkbox.isSelected())
        {
            Config.set("FilamentSensor", "show_intro", 0);
        }
    }

    private static String lookAndFeel()
    {
        String os_name = System.getProperty("os.name").toLowerCase();

        if (os_name.contains("win"))
        {
            return "com.sun.java.swing.plaf.windows.WindowsLookAndFeel";
        }
        /*else if (os_name.contains("linux"))
        {
            return "com.sun.java.swing.plaf.gtk.GTKLookAndFeel";
        }*/

        return UIManager.getSystemLookAndFeelClassName();
    }

    /**
     * Handle button-clicks and other ActionEvents sent by controls. Also see <a
     * href=
     * "http://docs.oracle.com/javase/1.4.2/docs/api/java/awt/event/ActionListener.html"
     * >here</a>
     *
     * @param e
     *            The ActionEvent to be handled.
     */
    @Override
    public void actionPerformed(ActionEvent e)
    {
        Object src = e.getSource();

        if (src == MainWindowMenu.f_exit)
        {
            frame.dispatchEvent(new WindowEvent(frame, WindowEvent.WINDOW_CLOSING));
        }
        else if (src == MainWindowMenu.f_new)
        {
            tabBar.setThumbnailTab(false);
        }
        else if (src == MainWindowMenu.f_open)
        {
            tabBar.setThumbnailTab(true);
        }
        else if (src == MainWindowMenu.f_csv)
        {
            tabBar.importCSV();
        }
        else if (src == MainWindowMenu.f_update)
        {
            DataProject.updateProject();
        }
        else if (src == MainWindowMenu.f_close)
        {
            tabBar.closeCurrentTab(false);
        }
        else if (src == MainWindowMenu.f_save)
        {
            tabBar.saveProject();
        }
        else if (src == MainWindowMenu.f_saveas)
        {
            tabBar.saveProject();
        }
        else if (src == MainWindowMenu.e_undo)
        {
            tabBar.getCurrentTab().undo();
        }
        else if (src == MainWindowMenu.e_redo)
        {
            tabBar.getCurrentTab().redo();
        }
        else if (src == MainWindowMenu.v_org)
        {
            tabBar.getCurrentTab().m_image_panel.resetView();
        }
        else if (src == MainWindowMenu.v_zoomin)
        {
            tabBar.getCurrentTab().m_image_panel.zoomIn();
        }
        else if (src == MainWindowMenu.v_zoomout)
        {
            tabBar.getCurrentTab().m_image_panel.zoomOut();
        }
        else if (src == MainWindowMenu.fl_delete)
        {
            tabBar.getCurrentTab().deleteSelected();
        }
        else if (src == MainWindowMenu.h_documentation)
        {
            JOptionPane.showMessageDialog(MainWindow.frame(), INFO, FilamentSensor.TITLE + " " +
                                          FilamentSensor.VERSION + " usage Information",
                                          JOptionPane.INFORMATION_MESSAGE, null);
        }
        else if (FilamentSensor.DEBUG)
        {
            System.out.println("Warning: unhandled ActionEvent " + e + "\nSource: " + src);
        }
    }

    /**
     * Saves things like window size and invokes config.saveConfig() on exit.
     *
     */
    @Override
    public void windowClosing(WindowEvent e)
    {
        if (tabBar.closeAll(false))
        {
            // save window properties and settings
            Dimension dim = frame.getSize();
            Config.set("MainWnd", "size_w", dim.width);
            Config.set("MainWnd", "size_h", dim.height);

            Point p = frame.getLocation();
            Config.set("MainWnd", "pos_w", p.x);
            Config.set("MainWnd", "pos_h", p.y);

            Config.saveConfig();

            e.getWindow().dispose();
        }
        else if (FilamentSensor.DEBUG)
        {
            System.out.println("Not ready yet, shutdown cancelled by user.");
        }
    }

    @Override
    public Object clone() throws CloneNotSupportedException
    {
        throw new CloneNotSupportedException();
    }

}
