package ui;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2014 Benjamin Eltzner
 *
 * FilTracker is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.*;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.TableRowSorter;

import core.AbstractFilament;
import core.Const;
import core.DataImage;
import core.DataParameters.Bin;
import core.DataParameters.Elog;
import core.DataParameters.Hough;
import core.DataParameters.Pre;
import core.DataParameters.Trace;
import core.DataProject;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

public class PickingView extends AbstractTab implements ActionListener, ChangeListener, ListSelectionListener,
PropertyChangeListener, TableModelListener

{

    private DataImage m_data;
    private static final Map<String, Color> m_color_map = makeColorMap();
    private static final Map<String, Integer> m_image_map = makeImageMap();

    // //// Picking view elements //////
    public JSplitPane m_picking_main_widget;

    // Left side
    private JTabbedPane m_left_pane;

    // Tab Preprocessing
    private JLabel m_image_quality;
    private JComboBox<String> m_spinbox_neighborhood;
    private JCheckBox m_checkbox_contrast, m_checkbox_singlepixel_pre, m_checkbox_laplace, m_checkbox_gauss,
                      m_checkbox_line_gauss, m_checkbox_preview, m_checkbox_force_defaults_pre, m_checkbox_cross;
    private SlideSpinPane m_slider_black, m_slider_white, m_slider_corr_size, m_slider_corr_zero;
    private SlideSpinPaneDouble m_slider_gauss_sigma_preprocessing, m_slider_laplace_factor, m_slider_line_gauss_sigma;
    private JRadioButton m_switch_preprocessing_gauss_first, m_switch_preprocessing_laplace_first;
    private JButton m_button_view_difference, m_button_default_parameters_pre;

    // Tab eLoG
    private SlideSpinPane m_slider_elog_threshold, m_slider_elog_orientations, m_slider_hough_orientations;
    private SlideSpinPaneDouble m_slider_elog_contrast, m_slider_elog_sigma_x, m_slider_elog_sigma_y;
    private JCheckBox m_checkbox_force_defaults_elog;
    private JRadioButton m_switch_elog_color, m_switch_elog_grey;
    private JButton m_button_elog, m_button_save_elog_image, m_button_save_elog_csv,
                    m_button_elog_batch, m_button_default_parameters_elog, m_button_hough_batch;

    // Tab Binarization/Tracing
    private JRadioButton m_switch_binarization_area, m_switch_binarization_rod;
    private SlideSpinPaneDouble m_slider_gauss_sigma_binarization, m_slider_area_significance,
                                m_slider_rod_significance;
    private SlideSpinPane m_slider_minimum_mean, m_slider_minimum_length, m_slider_minimum_angle,
                          m_slider_tolerance;
    private JCheckBox m_checkbox_thicken, m_checkbox_force_defaults_btc;
    private JButton m_button_widthmap, m_button_trace, m_button_batch_trace,
            m_button_default_parameters_btc;

    // Tab Filaments
    private TableModelFilaments m_table_model;
    private JTable m_filament_table;
    private JComboBox<String> m_spinbox_save_image, m_spinbox_filament_color;
    private JCheckBox m_checkbox_force_defaults_fil, m_checkbox_save_filaments;
    private JButton m_button_keep_all, m_button_clear, m_button_remedy_angles,
                    m_button_save_image, m_button_save_filaments_csv,
                    m_button_default_parameters_fil;

    // Right side
    public PanelImage m_image_panel;

    // Auxiliary
    private boolean m_lock = false;

    public PickingView(TabBar bar, String title)
    {
        // //////// Tab labels (project data) //////////
        super(new FlowLayout(FlowLayout.LEFT, 0, 0));
        m_tab_title = title;
        m_tab_bar = bar;

        this.add(new JLabel(IconCollection.TABICON));
        m_tab_label = new JLabel(title);
        m_tab_label.setBorder(BorderFactory.createEmptyBorder(1, 5, 0, 5));
        this.add(m_tab_label);

        this.add(new TabButton());
        setBorder(BorderFactory.createEmptyBorder(3, 3, 0, -1));
        setOpaque(false);

        // ////////////// Picking view ////////////////

        m_image_panel = new PanelImage();

        m_left_pane = new JTabbedPane();
        m_picking_main_widget = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, m_left_pane, m_image_panel);
        m_picking_main_widget.addPropertyChangeListener("dividerLocation", this);

        addTab(tabPre(), "Preprocessing",
               "Clean/denoise the image before the binarization process");
        addTab(tabBinTrace(), "Line Sensor",
               "Tweak the binarization and line tracing parameters");
        addTab(tabFilaments(), "Filaments", "View, edit and define filaments");
        addTab(tabElog(), "eLoG", "Set parameters for eLoG and Hough transformation.");

        m_left_pane.addChangeListener(this);
    }

    private static Map<String, Color> makeColorMap()
    {
        HashMap<String, Color> map = new HashMap<String, Color>();
        map.put("Orange", Color.orange);
        map.put("Dark Orange", new Color(255, 140, 0));
        map.put("Cyan", Color.cyan);
        map.put("Blue", Color.blue);
        return Collections.unmodifiableMap(map);
    }

    private static Map<String, Integer> makeImageMap()
    {
        HashMap<String, Integer> map = new HashMap<String, Integer>();
        map.put("Original", DataImage.ORIG);
        map.put("Preprocessed", DataImage.PRE);
        map.put("eLoG orientations", DataImage.ELOG_COLOR);
        map.put("eLoG greylevels", DataImage.ELOG_GREY);
        map.put("Binarized", DataImage.BIN);
        map.put("Line Sensor orientations", DataImage.ORI_MAP);
        return Collections.unmodifiableMap(map);
    }

    /**
     * This method sorts the string arrays for the spin boxes.
     */
    private static String[] sortedKeys(Map<String, ?> map)
    {
        if (!map.values().iterator().hasNext())
        {
            return null;
        }
        if (map.values().iterator().next() instanceof Number)
        {
            String[] strings = new String[map.size()];
            int counter = 0;
            for (int i = 0; counter < map.size(); i++)
            {
                for (String key : map.keySet())
                {
                    if (((Number) map.get(key)).intValue() == i)
                    {
                        strings[counter] = key;
                        counter++;
                    }
                }
            }
            return strings;
        }
        else
        {
            String[] strings = map.keySet().toArray(new String[0]);
            for(int i = 0; i < strings.length; i++)
            {
                for (int j=i+1 ; j < strings.length; j++)
                {
                    if(strings[i].compareTo(strings[j]) < 0)
                    {
                        String tmp = strings[j];
                        strings[j] = strings[i]; 
                        strings[i] = tmp;
                    }
                }
            }
            return strings;
        }
    }

    public static String stripExtension(String filename)
    {
        int i = filename.lastIndexOf('.');
        return ((i < 1 || i < filename.length() - 6) ? filename : filename.substring(0, i));
    }

    public boolean importCSV()
    {
        if (m_data == null)
        {
            return false;
        }
        return m_data.importCSV();
    }

    public void setData(DataImage data)
    {
        m_data = data;
        getControlValues();

        m_image_panel.setData(data);
        m_table_model.setData(data);
        m_image_quality.setText("Quality Score:  " + m_data.scoreImage());

        data.addChangeListener(this);
    }

    @Override
    protected void setTitle(String title)
    {
        m_tab_title = m_tab_bar.setTitle(this, title);
        m_tab_label.setText((m_data.hasChanged() ? "*" : "") + title);
    }

    public boolean showImage()
    {
        m_data.updatePreprocessing();
        return true;
    }

    public void deleteSelected()
    {
        if (m_left_pane.getSelectedIndex() != 2)
        {
            return;
        }

        m_data.deleteSelected();
    }

    public void undo()
    {
        m_data.undoChain();
    }

    public void redo()
    {
        m_data.redoChain();
    }

    private JPanel tabPre()
    {
        JPanel tab_preprocessing = new JPanel();
        tab_preprocessing.setLayout(new BoxLayout(tab_preprocessing, BoxLayout.Y_AXIS));

        tab_preprocessing.add(Box.createRigidArea(new Dimension(0, 15)));

        JPanel panel_score = new JPanel();
        panel_score.setLayout(new BoxLayout(panel_score, BoxLayout.X_AXIS));
        panel_score.setMaximumSize(new Dimension(400, 28));
        m_image_quality = new JLabel("", SwingConstants.LEFT);
        panel_score.add(m_image_quality);
        panel_score.add(Box.createHorizontalGlue());
        tab_preprocessing.add(panel_score);

        JPanel panel_contrast = new JPanel();
        panel_contrast.setLayout(new BoxLayout(panel_contrast, BoxLayout.X_AXIS));
        panel_contrast.setMaximumSize(new Dimension(400, 28));
        m_checkbox_contrast = new JCheckBox("Set contrast manually");
        m_checkbox_contrast.setSelected(false);
        m_checkbox_contrast.addActionListener(this);
        panel_contrast.add(m_checkbox_contrast);
        tab_preprocessing.add(panel_contrast);
        
        m_slider_black = new SlideSpinPane("Black value:", 0, 0, 255, 1, 0, 255);
        m_slider_black.setMaximumSize(new Dimension(400, 28));
        m_slider_black.setTickSpacing(4, 32);
        m_slider_black.setEnabled(false);
        m_slider_black.addActionListener(this, "updatePreprocessing");
        tab_preprocessing.add(m_slider_black);
        
        m_slider_white = new SlideSpinPane("White value:", 0, 0, 255, 1, 0, 255);
        m_slider_white.setMaximumSize(new Dimension(400, 28));
        m_slider_white.setTickSpacing(4, 32);
        m_slider_white.setEnabled(false);
        m_slider_white.addActionListener(this, "updatePreprocessing");
        tab_preprocessing.add(m_slider_white);

        tab_preprocessing.add(Box.createRigidArea(new Dimension(0, 5)));

        JPanel singlepixel_preprocessing = new JPanel();
        singlepixel_preprocessing.setLayout(new BoxLayout(singlepixel_preprocessing, BoxLayout.X_AXIS));
        singlepixel_preprocessing.setMaximumSize(new Dimension(400, 28));
        m_checkbox_singlepixel_pre = new JCheckBox("Remove standalone pixels");
        m_checkbox_singlepixel_pre.setSelected(true);
        m_checkbox_singlepixel_pre.addActionListener(this);
        singlepixel_preprocessing.add(m_checkbox_singlepixel_pre);
        tab_preprocessing.add(singlepixel_preprocessing);

        tab_preprocessing.add(Box.createVerticalGlue());

        JPanel jp_laplace = new JPanel();
        jp_laplace.setLayout(new BoxLayout(jp_laplace, BoxLayout.X_AXIS));
        jp_laplace.setMaximumSize(new Dimension(400, 28));
        m_checkbox_laplace = new JCheckBox("Laplace filter");
        m_checkbox_laplace.setSelected(false);
        m_checkbox_laplace.addActionListener(this);
        jp_laplace.add(m_checkbox_laplace);
        jp_laplace.add(Box.createHorizontalGlue());
        m_spinbox_neighborhood = new JComboBox<String>(new String[]{ "4 neighbor", "8 neighbor" });
        m_spinbox_neighborhood.setEditable(false);
        m_spinbox_neighborhood.setEnabled(false);
        m_spinbox_neighborhood.addActionListener(this);
        jp_laplace.add(m_spinbox_neighborhood);
        tab_preprocessing.add(jp_laplace);

        m_slider_laplace_factor = new SlideSpinPaneDouble("Factor:", .1, -25, 25, .1, 0, 5);
        m_slider_laplace_factor.setMaximumSize(new Dimension(400, 28));
        m_slider_laplace_factor.setTickSpacing(.1, 1.0);
        m_slider_laplace_factor.setEnabled(false);
        m_slider_laplace_factor.addActionListener(this, "updatePreprocessing");
        tab_preprocessing.add(m_slider_laplace_factor);

        tab_preprocessing.add(Box.createVerticalGlue());

        JPanel jp_gauss = new JPanel();
        jp_gauss.setLayout(new BoxLayout(jp_gauss, BoxLayout.X_AXIS));
        jp_gauss.setMaximumSize(new Dimension(400, 28));
        m_checkbox_gauss = new JCheckBox("Gaussian filter");
        m_checkbox_gauss.setSelected(false);
        m_checkbox_gauss.addActionListener(this);
        jp_gauss.add(m_checkbox_gauss);
        tab_preprocessing.add(jp_gauss);

        m_slider_gauss_sigma_preprocessing = new SlideSpinPaneDouble("Sigma:", 1.0, 0.0, 50.0, .1, 0.0, 15.0);
        m_slider_gauss_sigma_preprocessing.setMaximumSize(new Dimension(400, 28));
        m_slider_gauss_sigma_preprocessing.setTickSpacing(1.0, 5.0);
        m_slider_gauss_sigma_preprocessing.setEnabled(false);
        m_slider_gauss_sigma_preprocessing.addActionListener(this, "updatePreprocessing");
        tab_preprocessing.add(m_slider_gauss_sigma_preprocessing);

        tab_preprocessing.add(Box.createVerticalGlue());

        JPanel panel_cross = new JPanel();
        panel_cross.setLayout(new BoxLayout(panel_cross, BoxLayout.X_AXIS));
        panel_cross.setMaximumSize(new Dimension(400, 28));
        m_checkbox_cross = new JCheckBox("Cross Correlation filter");
        m_checkbox_cross.setSelected(false);
        m_checkbox_cross.addActionListener(this);
        panel_cross.add(m_checkbox_cross);
        tab_preprocessing.add(panel_cross);

        m_slider_corr_size = new SlideSpinPane("Size:", 1, 0, 10, 1, 0, 15);
        m_slider_corr_size.setMaximumSize(new Dimension(400, 28));
        m_slider_corr_size.setTickSpacing(1, 5);
        m_slider_corr_size.setEnabled(false);
        m_slider_corr_size.addActionListener(this, "updatePreprocessing");
        tab_preprocessing.add(m_slider_corr_size);

        m_slider_corr_zero = new SlideSpinPane("Zero (%):", 1, 0, 50, 1, 0, 100);
        m_slider_corr_zero.setMaximumSize(new Dimension(400, 28));
        m_slider_corr_zero.setTickSpacing(2, 10);
        m_slider_corr_zero.setEnabled(false);
        m_slider_corr_zero.addActionListener(this, "updatePreprocessing");
        tab_preprocessing.add(m_slider_corr_zero);

        tab_preprocessing.add(Box.createVerticalGlue());

        JPanel jp_line_gauss = new JPanel();
        jp_line_gauss.setLayout(new BoxLayout(jp_line_gauss, BoxLayout.X_AXIS));
        jp_line_gauss.setMaximumSize(new Dimension(400, 28));
        m_checkbox_line_gauss = new JCheckBox("Directed Gaussian filter");
        m_checkbox_line_gauss.setSelected(false);
        m_checkbox_line_gauss.addActionListener(this);
        jp_line_gauss.add(m_checkbox_line_gauss);
        tab_preprocessing.add(jp_line_gauss);

        m_slider_line_gauss_sigma = new SlideSpinPaneDouble("Sigma:", 1.0, 0.0, 50.0, .1, 0.0, 15.0);
        m_slider_line_gauss_sigma.setMaximumSize(new Dimension(400, 28));
        m_slider_line_gauss_sigma.setTickSpacing(1.0, 5.0);
        m_slider_line_gauss_sigma.setEnabled(false);
        m_slider_line_gauss_sigma.addActionListener(this, "updatePreprocessing");
        tab_preprocessing.add(m_slider_line_gauss_sigma);

        tab_preprocessing.add(Box.createVerticalGlue());

        JPanel jp_filter = new JPanel();
        jp_filter.setLayout(new BoxLayout(jp_filter, BoxLayout.X_AXIS));
        jp_filter.setBorder(BorderFactory.createTitledBorder("Image filter"));
        jp_filter.setMaximumSize(new Dimension(400, 28));
        m_switch_preprocessing_gauss_first = new JRadioButton("Gauss first");
        m_switch_preprocessing_gauss_first.setActionCommand("updatePreprocessing");
        m_switch_preprocessing_gauss_first.setSelected(true);
        m_switch_preprocessing_gauss_first.addActionListener(this);
        m_switch_preprocessing_laplace_first = new JRadioButton("Laplace first");
        m_switch_preprocessing_laplace_first.setActionCommand("updatePreprocessing");
        m_switch_preprocessing_laplace_first.setSelected(false);
        m_switch_preprocessing_laplace_first.addActionListener(this);
        ButtonGroup bg_filter = new ButtonGroup();
        bg_filter.add(m_switch_preprocessing_gauss_first);
        bg_filter.add(m_switch_preprocessing_laplace_first);
        jp_filter.add(m_switch_preprocessing_gauss_first);
        jp_filter.add(m_switch_preprocessing_laplace_first);
        tab_preprocessing.add(jp_filter);

        tab_preprocessing.add(Box.createVerticalGlue());

        JPanel jp_preview = new JPanel();
        jp_preview.setLayout(new BoxLayout(jp_preview, BoxLayout.X_AXIS));
        jp_preview.setMaximumSize(new Dimension(400, 28));
        jp_preview.add(Box.createHorizontalGlue());
        m_button_view_difference = new JButton("View filter delta");
        m_button_view_difference.addActionListener(this);
        jp_preview.add(m_button_view_difference);
        jp_preview.add(Box.createHorizontalGlue());
        m_checkbox_preview = new JCheckBox("Preview image result");
        m_checkbox_preview.setSelected(true);
        m_checkbox_preview.addActionListener(this);
        jp_preview.add(m_checkbox_preview);
        jp_preview.add(Box.createHorizontalGlue());
        tab_preprocessing.add(jp_preview);

        tab_preprocessing.add(Box.createVerticalGlue());

        JPanel defaults_pre = new JPanel();
        defaults_pre.setLayout(new BoxLayout(defaults_pre, BoxLayout.X_AXIS));
        defaults_pre.setMaximumSize(new Dimension(400, 28));
        defaults_pre.add(Box.createHorizontalGlue());
        m_button_default_parameters_pre = new JButton("Use parameters as default");
        m_button_default_parameters_pre.addActionListener(this);
        defaults_pre.add(m_button_default_parameters_pre);
        defaults_pre.add(Box.createHorizontalGlue());
        m_checkbox_force_defaults_pre = new JCheckBox("Enforce defaults");
        m_checkbox_force_defaults_pre.setSelected(false);
        m_checkbox_force_defaults_pre.addActionListener(this);
        defaults_pre.add(m_checkbox_force_defaults_pre);
        defaults_pre.add(Box.createHorizontalGlue());
        tab_preprocessing.add(defaults_pre);

        tab_preprocessing.add(Box.createRigidArea(new Dimension(0, 15)));

        return tab_preprocessing;
    }

    private JPanel tabBinTrace()
    {
        JPanel tab_bin_trace = new JPanel();
        tab_bin_trace.setLayout(new BoxLayout(tab_bin_trace, BoxLayout.Y_AXIS));

        tab_bin_trace.add(Box.createRigidArea(new Dimension(0, 15)));

        JPanel jp_method = new JPanel();
        jp_method.setLayout(new BoxLayout(jp_method, BoxLayout.X_AXIS));
        jp_method.setBorder(BorderFactory.createTitledBorder("Binarization method"));
        jp_method.setMaximumSize(new Dimension(400, 28));
        m_switch_binarization_area = new JRadioButton("Area");
        m_switch_binarization_area.setActionCommand("bin_method");
        m_switch_binarization_area.setSelected(true);
        m_switch_binarization_area.addActionListener(this);
        m_switch_binarization_rod = new JRadioButton("Directions");
        m_switch_binarization_rod.setActionCommand("bin_method");
        m_switch_binarization_rod.setSelected(false);
        m_switch_binarization_rod.addActionListener(this);        
        ButtonGroup bg_method = new ButtonGroup();
        bg_method.add(m_switch_binarization_area);
        bg_method.add(m_switch_binarization_rod);
        jp_method.add(m_switch_binarization_area);
        jp_method.add(m_switch_binarization_rod);
        tab_bin_trace.add(jp_method);

        tab_bin_trace.add(Box.createRigidArea(new Dimension(0, 15)));

        m_slider_minimum_mean = new SlideSpinPane("Minimum mean value:", 0, 0, 255, 1, 0, 150);
        m_slider_minimum_mean.setMaximumSize(new Dimension(400, 28));
        m_slider_minimum_mean.setTickSpacing(1, 25);
        m_slider_minimum_mean.addActionListener(this, "updateBinarization");
        tab_bin_trace.add(m_slider_minimum_mean);

        m_slider_gauss_sigma_binarization = new SlideSpinPaneDouble("Sigma:", 0.0, 0.0, 50.0, 0.1, 0.0, 20.0);
        m_slider_gauss_sigma_binarization.setMaximumSize(new Dimension(400, 28));
        m_slider_gauss_sigma_binarization.setTickSpacing(0.5, 5.0);
        m_slider_gauss_sigma_binarization.addActionListener(this, "updateBinarization");
        tab_bin_trace.add(m_slider_gauss_sigma_binarization);
        
        m_slider_area_significance = new SlideSpinPaneDouble("Minimum standard deviation:", 0.0, 0.0, 50.0, 0.1, 0.0, 20.0);
        m_slider_area_significance.setMaximumSize(new Dimension(400, 28));
        m_slider_area_significance.setVisible(false);
        m_slider_area_significance.setTickSpacing(1.0, 5.0);
        m_slider_area_significance.addActionListener(this, "updateBinarization");
        tab_bin_trace.add(m_slider_area_significance);
        
        m_slider_rod_significance = new SlideSpinPaneDouble("Minimum standard deviation:", 0.0, 0.0, 50.0, 0.1, 0.0, 20.0);
        m_slider_rod_significance.setMaximumSize(new Dimension(400, 28));
        m_slider_rod_significance.setTickSpacing(1.0, 5.0);
        m_slider_rod_significance.addActionListener(this, "updateBinarization");
        tab_bin_trace.add(m_slider_rod_significance);

        tab_bin_trace.add(Box.createRigidArea(new Dimension(0, 5)));

        JPanel panel_thicken = new JPanel();
        panel_thicken.setLayout(new BoxLayout(panel_thicken, BoxLayout.X_AXIS));
        panel_thicken.setMaximumSize(new Dimension(400, 28));
        m_checkbox_thicken = new JCheckBox("Thicken lines");
        m_checkbox_thicken.setSelected(false);
        m_checkbox_thicken.addActionListener(this);
        panel_thicken.add(m_checkbox_thicken);
        m_button_widthmap = new JButton("View width map");
        m_button_widthmap.addActionListener(this);
        panel_thicken.add(Box.createHorizontalGlue());
        panel_thicken.add(m_button_widthmap);
        panel_thicken.add(Box.createHorizontalGlue());
        tab_bin_trace.add(panel_thicken);

        tab_bin_trace.add(Box.createVerticalGlue());
        tab_bin_trace.add(Box.createRigidArea(new Dimension(0, 5)));
        tab_bin_trace.add(new JSeparator());
        tab_bin_trace.add(Box.createRigidArea(new Dimension(0, 5)));
        tab_bin_trace.add(Box.createVerticalGlue());

        // Filament sensor
        m_slider_minimum_length = new SlideSpinPane("Minimum filament length:", 1, 1, 400, 1, 0, 200);
        m_slider_minimum_length.setMaximumSize(new Dimension(400, 28));
        m_slider_minimum_length.setTickSpacing(5, 50);
        m_slider_minimum_length.addActionListener(this, "updateTraceParams");
        tab_bin_trace.add(m_slider_minimum_length);

        m_slider_minimum_angle = new SlideSpinPane("Minimum angle difference:", 0, 0, 90, 1);
        m_slider_minimum_angle.setMaximumSize(new Dimension(400, 28));
        m_slider_minimum_angle.setTickSpacing(1, 15);
        m_slider_minimum_angle.addActionListener(this, "updateTraceParams");
        tab_bin_trace.add(m_slider_minimum_angle);

        m_slider_tolerance = new SlideSpinPane("Tolerance (%):", 5, 0, 50, 1);
        m_slider_tolerance.setMaximumSize(new Dimension(400, 28));
        m_slider_tolerance.setTickSpacing(1, 10);
        m_slider_tolerance.addActionListener(this, "updateTraceParams");
        tab_bin_trace.add(m_slider_tolerance);

        tab_bin_trace.add(Box.createVerticalGlue());

        JPanel jp_tracebuttons = new JPanel();
        jp_tracebuttons.setLayout(new BoxLayout(jp_tracebuttons, BoxLayout.X_AXIS));
        jp_tracebuttons.setMaximumSize(new Dimension(400, 28));
        jp_tracebuttons.add(Box.createHorizontalGlue());
        m_button_trace = new JButton("Trace filaments");
        m_button_trace.addActionListener(this);
        jp_tracebuttons.add(m_button_trace);
        jp_tracebuttons.add(Box.createHorizontalGlue());
        m_button_batch_trace = new JButton("Batch Tracing");
        m_button_batch_trace.addActionListener(this);
        jp_tracebuttons.add(m_button_batch_trace);
        jp_tracebuttons.add(Box.createHorizontalGlue());
        tab_bin_trace.add(jp_tracebuttons);

        tab_bin_trace.add(Box.createVerticalGlue());
        tab_bin_trace.add(Box.createRigidArea(new Dimension(0, 15)));

        JPanel defaults_btc = new JPanel();
        defaults_btc.setLayout(new BoxLayout(defaults_btc, BoxLayout.X_AXIS));
        defaults_btc.setMaximumSize(new Dimension(400, 28));
        defaults_btc.add(Box.createHorizontalGlue());
        m_button_default_parameters_btc = new JButton("Use parameters as default");
        m_button_default_parameters_btc.addActionListener(this);
        defaults_btc.add(m_button_default_parameters_btc);
        defaults_btc.add(Box.createHorizontalGlue());
        m_checkbox_force_defaults_btc = new JCheckBox("Enforce defaults");
        m_checkbox_force_defaults_btc.setSelected(false);
        m_checkbox_force_defaults_btc.addActionListener(this);
        defaults_btc.add(m_checkbox_force_defaults_btc);
        defaults_btc.add(Box.createHorizontalGlue());
        tab_bin_trace.add(defaults_btc);

        tab_bin_trace.add(Box.createRigidArea(new Dimension(0, 15)));

        return tab_bin_trace;
    }

    private JPanel tabFilaments()
    {
        JPanel tab_filaments = new JPanel();
        tab_filaments.setLayout(new BoxLayout(tab_filaments, BoxLayout.Y_AXIS));

        tab_filaments.add(Box.createRigidArea(new Dimension(0, 15)));

        // List of filaments
        m_table_model = new TableModelFilaments();
        m_filament_table = new JTable(m_table_model);
        m_filament_table.setRowSorter(new TableRowSorter<TableModelFilaments>(m_table_model));
        m_filament_table.setPreferredScrollableViewportSize(new Dimension(350, 200));
        m_filament_table.setFillsViewportHeight(true);
        m_filament_table.getSelectionModel().addListSelectionListener(this);
        JScrollPane jsp_filtable = new JScrollPane(m_filament_table);
        jsp_filtable.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
        tab_filaments.add(jsp_filtable);
        m_table_model.addTableModelListener(this);

        tab_filaments.add(Box.createRigidArea(new Dimension(0, 10)));
        JPanel panel_clear = new JPanel();
        panel_clear.setLayout(new BoxLayout(panel_clear, BoxLayout.X_AXIS));
        panel_clear.setMaximumSize(new Dimension(400, 28));
        panel_clear.add(Box.createRigidArea(new Dimension(10, 0)));
        m_button_keep_all = new JButton("Discard/Keep all");
        m_button_keep_all.addActionListener(this);
        panel_clear.add(m_button_keep_all);
        panel_clear.add(Box.createHorizontalGlue());
        m_button_clear = new JButton("Remove discarded");
        m_button_clear.addActionListener(this);
        panel_clear.add(m_button_clear);
        panel_clear.add(Box.createRigidArea(new Dimension(10, 0)));
        tab_filaments.add(panel_clear);

        tab_filaments.add(Box.createRigidArea(new Dimension(0, 10)));
        JPanel panel_remedy = new JPanel();
        panel_remedy.setLayout(new BoxLayout(panel_remedy, BoxLayout.X_AXIS));
        panel_remedy.setMaximumSize(new Dimension(400, 28));
        panel_remedy.add(Box.createHorizontalGlue());
        m_button_remedy_angles = new JButton("Remedy Orientations");
        m_button_remedy_angles.addActionListener(this);
        panel_remedy.add(m_button_remedy_angles);
        panel_remedy.add(Box.createRigidArea(new Dimension(10, 0)));
        tab_filaments.add(panel_remedy);

        tab_filaments.add(Box.createVerticalGlue());

        JPanel panel_save = new JPanel();
        panel_save.setLayout(new BoxLayout(panel_save, BoxLayout.X_AXIS));
        panel_save.setMaximumSize(new Dimension(400, 28));
        panel_save.add(Box.createRigidArea(new Dimension(10, 0)));
        m_spinbox_save_image = new JComboBox<String>(sortedKeys(m_image_map));
        m_spinbox_save_image.setEditable(false);
        panel_save.add(m_spinbox_save_image);
        panel_save.add(Box.createHorizontalGlue());
        tab_filaments.add(panel_save);

        tab_filaments.add(Box.createRigidArea(new Dimension(0, 15)));

        JPanel panel_color = new JPanel();
        panel_color.setLayout(new BoxLayout(panel_color, BoxLayout.X_AXIS));
        panel_color.setMaximumSize(new Dimension(400, 28));
        panel_color.add(Box.createRigidArea(new Dimension(10, 0)));
        m_checkbox_save_filaments = new JCheckBox("Include filaments");
        m_checkbox_save_filaments.setSelected(false);
        m_checkbox_save_filaments.addActionListener(this);
        panel_color.add(m_checkbox_save_filaments);
        panel_color.add(Box.createHorizontalGlue());
        m_spinbox_filament_color = new JComboBox<String>(sortedKeys(m_color_map));
        m_spinbox_filament_color.setEditable(false);
        m_spinbox_filament_color.setVisible(false);
        m_spinbox_filament_color.setSize(m_spinbox_save_image.getSize());
        panel_color.add(m_spinbox_filament_color);
        panel_color.add(Box.createHorizontalGlue());
        tab_filaments.add(panel_color);

        tab_filaments.add(Box.createRigidArea(new Dimension(0, 15)));

        JPanel panel_save_image = new JPanel();
        panel_save_image.setLayout(new BoxLayout(panel_save_image, BoxLayout.X_AXIS));
        panel_save_image.setMaximumSize(new Dimension(400, 28));
        panel_save_image.add(Box.createRigidArea(new Dimension(10, 0)));
        m_button_save_image = new JButton("Save Image");
        m_button_save_image.addActionListener(this);
        panel_save_image.add(m_button_save_image);
        panel_save_image.add(Box.createHorizontalGlue());
        tab_filaments.add(panel_save_image);

        tab_filaments.add(Box.createVerticalGlue());

        JPanel jp_export = new JPanel();
        jp_export.add(Box.createHorizontalGlue());
        jp_export.setLayout(new BoxLayout(jp_export, BoxLayout.X_AXIS));
        jp_export.setMaximumSize(new Dimension(400, 28));
        jp_export.add(Box.createRigidArea(new Dimension(10, 0)));
        m_button_save_filaments_csv = new JButton("Export filaments as .csv");
        m_button_save_filaments_csv.addActionListener(this);
        jp_export.add(m_button_save_filaments_csv);
        jp_export.add(Box.createHorizontalGlue());
        tab_filaments.add(jp_export);

        tab_filaments.add(Box.createVerticalGlue());

        JPanel defaults_fil = new JPanel();
        defaults_fil.setLayout(new BoxLayout(defaults_fil, BoxLayout.X_AXIS));
        defaults_fil.setMaximumSize(new Dimension(400, 28));
        defaults_fil.add(Box.createHorizontalGlue());
        m_button_default_parameters_fil = new JButton("Use parameters as default");
        m_button_default_parameters_fil.addActionListener(this);
        defaults_fil.add(m_button_default_parameters_fil);
        defaults_fil.add(Box.createHorizontalGlue());
        m_checkbox_force_defaults_fil = new JCheckBox("Enforce defaults");
        m_checkbox_force_defaults_fil.setSelected(false);
        m_checkbox_force_defaults_fil.addActionListener(this);
        defaults_fil.add(m_checkbox_force_defaults_fil);
        defaults_fil.add(Box.createHorizontalGlue());
        tab_filaments.add(defaults_fil);

        tab_filaments.add(Box.createRigidArea(new Dimension(0, 15)));

        return tab_filaments;
    }

    private JPanel tabElog()
    {
        JPanel tab_elog = new JPanel();
        tab_elog.setLayout(new BoxLayout(tab_elog, BoxLayout.Y_AXIS));

        tab_elog.add(Box.createRigidArea(new Dimension(0, 15)));

        JPanel switch_elog = new JPanel();
        switch_elog.setLayout(new BoxLayout(switch_elog, BoxLayout.X_AXIS));
        switch_elog.setBorder(BorderFactory.createTitledBorder("Select Image"));
        switch_elog.setMaximumSize(new Dimension(400, 28));
        m_switch_elog_color = new JRadioButton("Color Image");
        m_switch_elog_color.setActionCommand("updateELoGImage");
        m_switch_elog_color.setSelected(true);
        m_switch_elog_color.addActionListener(this);
        m_switch_elog_grey = new JRadioButton("Greyscale Image");
        m_switch_elog_grey.setActionCommand("updateELoGImage");
        m_switch_elog_grey.setSelected(false);
        m_switch_elog_grey.addActionListener(this);
        ButtonGroup switch_group_elog = new ButtonGroup();
        switch_group_elog.add(m_switch_elog_color);
        switch_group_elog.add(m_switch_elog_grey);
        switch_elog.add(m_switch_elog_color);
        switch_elog.add(m_switch_elog_grey);
        tab_elog.add(switch_elog);

        tab_elog.add(Box.createRigidArea(new Dimension(0, 15)));

        m_slider_elog_threshold = new SlideSpinPane("Brightness threshold:", 0, 0, 255, 1, 0, 100);
        m_slider_elog_threshold.setMaximumSize(new Dimension(400, 28));
        m_slider_elog_threshold.setTickSpacing(5, 25);
        m_slider_elog_threshold.addActionListener(this, "updateELoGParams");
        tab_elog.add(m_slider_elog_threshold);
        
        m_slider_elog_orientations = new SlideSpinPane("Orientations:", 0, 0, 200, 1, 0, 180);
        m_slider_elog_orientations.setMaximumSize(new Dimension(400, 28));
        m_slider_elog_orientations.setTickSpacing(5, 30);
        m_slider_elog_orientations.addActionListener(this, "updateELoGParams");
        tab_elog.add(m_slider_elog_orientations);

        m_slider_elog_contrast = new SlideSpinPaneDouble("Contrast:", 0.0, 0.0, 30.0, 0.1, 0.0, 20.0);
        m_slider_elog_contrast.setMaximumSize(new Dimension(400, 28));
        m_slider_elog_contrast.setTickSpacing(0.5, 5.0);
        m_slider_elog_contrast.addActionListener(this, "updateELoGParams");
        tab_elog.add(m_slider_elog_contrast);

        tab_elog.add(Box.createRigidArea(new Dimension(0, 5)));

        m_slider_elog_sigma_x = new SlideSpinPaneDouble("Horizontal Sigma:", 0.0, 0.0, 50.0, 0.1, 0.0, 20.0);
        m_slider_elog_sigma_x.setMaximumSize(new Dimension(400, 28));
        m_slider_elog_sigma_x.setTickSpacing(0.5, 5.0);
        m_slider_elog_sigma_x.addActionListener(this, "updateELoGParams");
        tab_elog.add(m_slider_elog_sigma_x);

        m_slider_elog_sigma_y = new SlideSpinPaneDouble("Vertical Sigma:", 0.0, 0.0, 50.0, 0.1, 0.0, 20.0);
        m_slider_elog_sigma_y.setMaximumSize(new Dimension(400, 28));
        m_slider_elog_sigma_y.setTickSpacing(0.5, 5.0);
        m_slider_elog_sigma_y.addActionListener(this, "updateELoGParams");
        tab_elog.add(m_slider_elog_sigma_y);

        tab_elog.add(Box.createVerticalGlue());

        JPanel jp_elog = new JPanel();
        jp_elog.setLayout(new BoxLayout(jp_elog, BoxLayout.X_AXIS));
        jp_elog.setMaximumSize(new Dimension(400, 28));
        m_button_save_elog_image = new JButton("Save eLoG Image");
        m_button_save_elog_image.addActionListener(this);
        m_button_elog = new JButton("   eLoGs   ");
        m_button_elog.addActionListener(this);
        jp_elog.add(Box.createHorizontalGlue());
        jp_elog.add(m_button_save_elog_image);
        jp_elog.add(Box.createHorizontalGlue());
        jp_elog.add(m_button_elog);
        jp_elog.add(Box.createHorizontalGlue());
        tab_elog.add(jp_elog);

        tab_elog.add(Box.createRigidArea(new Dimension(0, 15)));

        JPanel panel_elog_export = new JPanel();
        panel_elog_export.setLayout(new BoxLayout(panel_elog_export, BoxLayout.X_AXIS));
        panel_elog_export.setMaximumSize(new Dimension(400, 28));
        m_button_save_elog_csv = new JButton("Export histogram");
        m_button_save_elog_csv.addActionListener(this);
        m_button_elog_batch = new JButton("eLoG batch");
        m_button_elog_batch.addActionListener(this);
        panel_elog_export.add(Box.createHorizontalGlue());
        panel_elog_export.add(m_button_save_elog_csv);
        panel_elog_export.add(Box.createHorizontalGlue());
        panel_elog_export.add(m_button_elog_batch);
        panel_elog_export.add(Box.createHorizontalGlue());
        tab_elog.add(panel_elog_export);

        tab_elog.add(Box.createVerticalGlue());
        
        m_slider_hough_orientations = new SlideSpinPane("Orientations:", 0, 0, 360, 1, 0, 500);
        m_slider_hough_orientations.setMaximumSize(new Dimension(400, 28));
        m_slider_hough_orientations.setTickSpacing(10, 60);
        m_slider_hough_orientations.addActionListener(this, "updateELoGParams");
        tab_elog.add(m_slider_hough_orientations);
        
        JPanel panel_hough = new JPanel();
        panel_hough.setLayout(new BoxLayout(panel_hough, BoxLayout.X_AXIS));
        panel_hough.setMaximumSize(new Dimension(400, 28));
        m_button_hough_batch = new JButton("Hough batch");
        m_button_hough_batch.addActionListener(this);
        panel_hough.add(Box.createHorizontalGlue());
        panel_hough.add(m_button_hough_batch);
        panel_hough.add(Box.createHorizontalGlue());
        tab_elog.add(panel_hough);

        tab_elog.add(Box.createVerticalGlue());

        JPanel defaults_elog = new JPanel();
        defaults_elog.setLayout(new BoxLayout(defaults_elog, BoxLayout.X_AXIS));
        defaults_elog.setMaximumSize(new Dimension(400, 28));
        defaults_elog.add(Box.createHorizontalGlue());
        m_button_default_parameters_elog = new JButton("Use parameters as default");
        m_button_default_parameters_elog.addActionListener(this);
        defaults_elog.add(m_button_default_parameters_elog);
        defaults_elog.add(Box.createHorizontalGlue());
        m_checkbox_force_defaults_elog = new JCheckBox("Enforce defaults");
        m_checkbox_force_defaults_elog.setSelected(false);
        m_checkbox_force_defaults_elog.addActionListener(this);
        defaults_elog.add(m_checkbox_force_defaults_elog);
        defaults_elog.add(Box.createHorizontalGlue());
        tab_elog.add(defaults_elog);

        tab_elog.add(Box.createRigidArea(new Dimension(0, 15)));

        return tab_elog;
    }

    private void addTab(JPanel panel, String name, String tooltip)
    {
        JScrollPane scroller = new JScrollPane(panel, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
                ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
        scroller.getVerticalScrollBar().setUnitIncrement(Const.SCROLL);
        m_left_pane.addTab(name, null, scroller, tooltip);
    }

    private void getControlValues()
    {
        // prevent update* methods from overwriting the values we want to read
        m_lock = true;
        m_checkbox_contrast.setSelected(m_data.getParameter(Pre.contrast) != 0);
        m_slider_black.setValue(m_data.getBlackWhite(false));
        m_slider_white.setValue(m_data.getBlackWhite(true));
        m_checkbox_singlepixel_pre.setSelected(m_data.getParameter(Pre.rm_sng_px) != 0);
        m_checkbox_laplace.setSelected(m_data.getParameter(Pre.do_laplace) != 0);
        m_slider_laplace_factor.setEnabled(m_data.getParameter(Pre.do_laplace) != 0);
        m_spinbox_neighborhood.setEnabled(m_data.getParameter(Pre.do_laplace) != 0);
        m_slider_laplace_factor.setValue(m_data.getParameter(Pre.lpfac) / 10.0);
        m_spinbox_neighborhood.setSelectedIndex(m_data.getParameter(Pre.lpmask));
        m_checkbox_gauss.setSelected(m_data.getParameter(Pre.do_gauss) != 0);
        m_slider_gauss_sigma_preprocessing.setEnabled(m_data.getParameter(Pre.do_gauss) != 0);
        m_slider_gauss_sigma_preprocessing.setValue(m_data.getParameter(Pre.fsigma) / 10.0);
        m_switch_preprocessing_gauss_first.setSelected(m_data.getParameter(Pre.order) == 0);
        m_checkbox_line_gauss.setSelected(m_data.getParameter(Pre.do_line_gauss) != 0);
        m_slider_line_gauss_sigma.setEnabled(m_data.getParameter(Pre.do_line_gauss) != 0);
        m_slider_line_gauss_sigma.setValue(m_data.getParameter(Pre.line_sigma) / 10.0);
        m_checkbox_cross.setSelected(m_data.getParameter(Pre.cross_corr) != 0);
        m_slider_corr_size.setEnabled(m_data.getParameter(Pre.cross_corr) != 0);
        m_slider_corr_size.setValue(m_data.getParameter(Pre.corr_mask_size));
        m_slider_corr_zero.setEnabled(m_data.getParameter(Pre.cross_corr) != 0);
        m_slider_corr_zero.setValue(m_data.getParameter(Pre.corr_zero));

        m_checkbox_preview.setSelected(m_data.getParameter(Pre.preview) != 0);

        m_switch_binarization_rod.setSelected(m_data.getParameter(Bin.method) == 1);
        m_slider_area_significance.setVisible(m_switch_binarization_area.isSelected());
        m_slider_rod_significance.setVisible(m_switch_binarization_rod.isSelected());
        m_slider_minimum_mean.setValue(m_data.getParameter(Bin.minmean));
        m_slider_gauss_sigma_binarization.setValue(m_data.getParameter(Bin.sigma) / 10.0);
        m_slider_area_significance.setValue(m_data.getParameter(Bin.area_significance) / 10.0);
        m_slider_rod_significance.setValue(m_data.getParameter(Bin.rod_significance) / 10.0);
        m_checkbox_thicken.setSelected(m_data.getParameter(Bin.thicken) != 0);

        m_slider_elog_threshold.setValue(m_data.getElogThreshold());
        m_slider_elog_orientations.setValue(m_data.getParameter(Elog.orientations));
        m_slider_elog_contrast.setValue(m_data.getParameter(Elog.contrast) / 10.0);
        m_slider_elog_sigma_x.setValue(m_data.getParameter(Elog.sigma_x) / 10.0);
        m_slider_elog_sigma_y.setValue(m_data.getParameter(Elog.sigma_y) / 10.0);

        m_slider_hough_orientations.setValue(m_data.getParameter(Hough.orientations));

        m_slider_minimum_length.setValue(m_data.getParameter(Trace.minlen));
        m_slider_minimum_angle.setValue(m_data.getParameter(Trace.minangle));
        m_slider_tolerance.setValue(m_data.getParameter(Trace.tolerance));
        m_lock = false;
    }

    private void updatePreprocessing()
    {
        if (m_lock)
        {
            return;
        }

        m_data.setParameter(Pre.contrast, m_checkbox_contrast.isSelected() ? 1 : 0);
        m_data.setParameter(Pre.black, m_checkbox_contrast.isSelected() ? m_slider_black.getValue() : -1);
        m_data.setParameter(Pre.white, m_checkbox_contrast.isSelected() ? m_slider_white.getValue() : -1);
        m_data.setParameter(Pre.rm_sng_px, m_checkbox_singlepixel_pre.isSelected() ? 1 : 0);
        m_data.setParameter(Pre.do_laplace, m_checkbox_laplace.isSelected() ? 1 : 0);
        m_data.setParameter(Pre.lpfac,
                Math.round(((Double) (m_slider_laplace_factor.getValue() * 10)).floatValue()));
        m_data.setParameter(Pre.lpmask, m_spinbox_neighborhood.getSelectedIndex());
        m_data.setParameter(Pre.do_gauss, m_checkbox_gauss.isSelected() ? 1 : 0);
        m_data.setParameter(Pre.fsigma,
                Math.round(((Double) (m_slider_gauss_sigma_preprocessing.getValue() * 10)).floatValue()));
        m_data.setParameter(Pre.order, m_switch_preprocessing_laplace_first.isSelected() ? 1 : 0);
        m_data.setParameter(Pre.do_line_gauss, m_checkbox_line_gauss.isSelected() ? 1 : 0);
        m_data.setParameter(Pre.line_sigma,
                Math.round(((Double) (m_slider_line_gauss_sigma.getValue() * 10)).floatValue()));
        m_data.setParameter(Pre.cross_corr, m_checkbox_cross.isSelected() ? 1 : 0);
        m_data.setParameter(Pre.corr_mask_size, m_slider_corr_size.getValue());
        m_data.setParameter(Pre.corr_zero, m_slider_corr_zero.getValue());
        
        m_data.setParameter(Pre.preview, m_checkbox_preview.isSelected() ? 1 : 0);

        m_data.updatePreprocessing();
    }

    private void updateELoGParams()
    {
        if (m_lock)
        {
            return;
        }

        m_data.setParameter(Elog.image, (m_switch_elog_color.isSelected() ? 0 : 1));
        m_data.setParameter(Elog.threshold, m_slider_elog_threshold.getValue());
        m_data.setParameter(Elog.orientations, m_slider_elog_orientations.getValue());
        m_data.setParameter(Elog.contrast, Math.round(((Double) (m_slider_elog_contrast.getValue() * 10)).floatValue()));
        m_data.setParameter(Elog.sigma_x, Math.round(((Double) (m_slider_elog_sigma_x.getValue() * 10)).floatValue()));
        m_data.setParameter(Elog.sigma_y, Math.round(((Double) (m_slider_elog_sigma_y.getValue() * 10)).floatValue()));

        m_data.setParameter(Hough.orientations, m_slider_hough_orientations.getValue());
    }

    private void updateBinarization()
    {
        if (m_lock)
        {
            return;
        }

        m_data.setParameter(Bin.method, (m_switch_binarization_area.isSelected() ? 0
                : (m_switch_binarization_rod.isSelected() ? 1 : 2)));
        m_data.setParameter(Bin.minmean, m_slider_minimum_mean.getValue());
        m_data.setParameter(Bin.sigma,
                Math.round(((Double) (m_slider_gauss_sigma_binarization.getValue() * 10)).floatValue()));
        m_data.setParameter(Bin.area_significance,
                Math.round(((Double) (m_slider_area_significance.getValue() * 10)).floatValue()));
        m_data.setParameter(Bin.rod_significance,
                Math.round(((Double) (m_slider_rod_significance.getValue() * 10)).floatValue()));
        m_data.setParameter(Bin.thicken, m_checkbox_thicken.isSelected() ? 1 : 0);

        m_data.updateBinarization();
    }

    private void updateTraceParams()
    {
        if (m_lock)
        {
            return;
        }

        m_data.setParameter(Trace.minlen, m_slider_minimum_length.getValue());
        m_data.setParameter(Trace.minangle, m_slider_minimum_angle.getValue());
        m_data.setParameter(Trace.tolerance, m_slider_tolerance.getValue());
        m_data.updateTraceParams();
    }

    @Override
    public void actionPerformed(ActionEvent e)
    {
        if (m_data == null)
        {
            return;
        }

        Object src = e.getSource();
        String com = e.getActionCommand();

        if (com.equals("bin_method"))
        {
            m_slider_area_significance.setVisible(m_switch_binarization_area.isSelected());
            m_slider_rod_significance.setVisible(m_switch_binarization_rod.isSelected());
            updateBinarization();
        }
        else if (src == m_checkbox_contrast)
        {
            m_slider_black.setEnabled(m_checkbox_contrast.isSelected());
            m_slider_white.setEnabled(m_checkbox_contrast.isSelected());
            updatePreprocessing();
        }
        else if (src == m_checkbox_laplace || src == m_checkbox_gauss)
        {
            m_slider_laplace_factor.setEnabled(m_checkbox_laplace.isSelected());
            m_spinbox_neighborhood.setEnabled(m_checkbox_laplace.isSelected());
            m_slider_gauss_sigma_preprocessing.setEnabled(m_checkbox_gauss.isSelected());
            m_switch_preprocessing_gauss_first.setEnabled(m_checkbox_gauss.isSelected()
                    && m_checkbox_laplace.isSelected());
            m_switch_preprocessing_laplace_first.setEnabled(m_checkbox_gauss.isSelected()
                    && m_checkbox_laplace.isSelected());

            updatePreprocessing();
        }
        else if (src == m_checkbox_line_gauss)
        {
            m_slider_line_gauss_sigma.setEnabled(m_checkbox_line_gauss.isSelected());
            updatePreprocessing();
        }
        else if (src == m_checkbox_cross)
        {
            m_slider_corr_size.setEnabled(m_checkbox_cross.isSelected());
            m_slider_corr_zero.setEnabled(m_checkbox_cross.isSelected());
            updatePreprocessing();
        }
        else if (src == m_checkbox_singlepixel_pre || src == m_spinbox_neighborhood ||
                com.equals("updatePreprocessing"))
        {
            updatePreprocessing();
        }
        else if (src == m_checkbox_preview)
        {
            m_data.setParameter(Pre.preview, m_checkbox_preview.isSelected() ? 1 : 0);
            m_image_panel.update(m_left_pane.getSelectedIndex());
        }
        else if (com.equals("updateELoGParams"))
        {
            updateELoGParams();
        }
        else if (com.equals("updateELoGImage"))
        {
            updateELoGParams();
            m_image_panel.update(m_left_pane.getSelectedIndex());
        }
        else if (src == m_button_elog)
        {
            updateELoGParams();
            m_data.calculateELoGs();
            m_image_panel.update(m_left_pane.getSelectedIndex());
        }
        else if (src == m_button_save_elog_image)
        {
            m_data.saveImage(null, "eLoG", DataImage.ELOG_COLOR, false, null);
        }
        else if (src == m_button_save_elog_csv)
        {
            m_data.exportHistogram(null, true, "", DataImage.ELOG_COLOR);
        }
        else if (src == m_button_elog_batch)
        {
            DataProject.batchElog(m_data);
        }
        else if (src == m_button_hough_batch)
        {
            DataProject.batchHough(m_data);
        }
        else if (src == m_checkbox_thicken || com.equals("updateBinarization"))
        {
            updateBinarization();
        }
        else if (src == m_button_widthmap)
        {
            new PopUpWidthMap(m_data);
        }
        else if (com.equals("updateTraceParams"))
        {
            updateTraceParams();
        }
        else if (src == m_button_batch_trace)
        {
            DataProject.batchProcessing(m_data);
            m_tab_bar.saveProject();
        }
        else if (src == m_button_trace)
        {
            m_data.scanFilaments();
        }
        else if (src == m_button_view_difference)
        {
            new PopUpFilterDiff(m_data);
        }
        else if (src == m_button_default_parameters_pre || src == m_button_default_parameters_elog
                || src == m_button_default_parameters_btc || src == m_button_default_parameters_fil)
        {
            if (FilamentSensor.DEBUG)
            {
                System.out.println("Setting default parameters.");
            }

            DataProject.setForce(false);
            DataProject.setDefaultParameters(m_data.parameters());
            if (m_checkbox_force_defaults_pre.isSelected())
            {
                DataProject.setForce(true);
            }
        }
        else if (src == m_checkbox_force_defaults_pre || src == m_checkbox_force_defaults_elog
                || src == m_checkbox_force_defaults_btc || src == m_checkbox_force_defaults_fil)
        {
            System.out.println("Setting checkboxes.");
            boolean force = ((JCheckBox) src).isSelected();
            m_checkbox_force_defaults_pre.setSelected(force);
            m_checkbox_force_defaults_elog.setSelected(force);
            m_checkbox_force_defaults_btc.setSelected(force);
            m_checkbox_force_defaults_fil.setSelected(force);
        }
        else if (src == m_button_clear)
        {
            m_data.cleanTracedFilaments();
            m_table_model.fireTableDataChanged();
        }
        else if (src == m_button_keep_all)
        {
            m_data.keepDiscardAllFilaments();
            m_table_model.fireTableDataChanged();
        }
        else if (src == m_button_remedy_angles)
        {
            m_data.remedyOldAngles();
            m_table_model.fireTableDataChanged();
        }
        else if (src == m_checkbox_save_filaments)
        {
            m_spinbox_filament_color.setVisible(m_checkbox_save_filaments.isSelected());
        }
        else if (src == m_button_save_image)
        {
            Color color = Color.orange;
            if(m_checkbox_save_filaments.isSelected())
            {
                color = m_color_map.get((String) m_spinbox_filament_color.getSelectedItem());
            }
            m_data.saveImage(null, null, m_image_map.get((String) m_spinbox_save_image.getSelectedItem()),
                             m_checkbox_save_filaments.isSelected(), color);
        }
        else if (src == m_button_save_filaments_csv)
        {
            m_data.exportCSV(null, false, true);
        }
        else if (FilamentSensor.DEBUG)
        {
            System.out.println("Warning: unhandled ActionEvent " + e + "\nSource: " + src);
        }
    }

    /**
     * Handles tab-changes from m_left_pane
     * ("Preprocessing, Binarization/Tracing, Filaments") and updates the
     * ImagePanel, filament table and tab title (to indicate unsaved changes)
     * when DataImage was modified.
     *
     * @param e
     *            the ChangeEvent (TODO: link?)
     */
    @Override
    public void stateChanged(ChangeEvent e)
    {
        if (m_data == null)
        {
            return;
        }

        // update change indicator (*title)
        setTitle(m_tab_title);

        m_lock = true;
        m_table_model.fireTableDataChanged();

        int min = Integer.MAX_VALUE, max = 0;
        for (AbstractFilament fil : m_data.getSelection())
        {
            int row = m_data.getRowOfEntry(fil);

            // If FilamentChains have changed and fil is not found.
            if (row == -1)
            {
                max = -1;
                break;
            }

            int i = m_filament_table.convertRowIndexToView(row);
            m_filament_table.addRowSelectionInterval(i, i);

            if (i < min)
            {
                min = i;
            }
            if (i > max)
            {
                max = i;
            }
        }

        if (max == -1)
        {
            m_data.clearSelection();
            max = 0;
        }

        // scroll to newly selected rows
        if (min < Integer.MAX_VALUE && min != m_table_model.scroll_min)
        {
            m_table_model.scroll_min = min;
            m_filament_table.scrollRectToVisible(m_filament_table.getCellRect(min, 0, true));
        }
        if (max > 0 && max != m_table_model.scroll_max)
        {
            m_table_model.scroll_max = max;
            m_filament_table.scrollRectToVisible(m_filament_table.getCellRect(max, 0, true));
        }

        m_lock = false;

        // TODO: This is inflexible.
        if (e.getSource() == m_left_pane && m_left_pane.getSelectedIndex() == 1)
        {
            m_data.updateBinarization();
        }

        m_image_panel.update(m_left_pane.getSelectedIndex());
        m_tab_bar.updateMenu(e);
    }

    @Override
    public void tableChanged(TableModelEvent e)
    {
        Object src = e.getSource();

        if (src == m_table_model)
        {
            m_image_panel.repaint();
        }
    }

    /**
     * Handles ListSelectionEvents sent when the selection of the
     * filaments-JTable changed. Updates data.selection
     */
    @Override
    public void valueChanged(ListSelectionEvent e)
    {
        ListSelectionModel model = (ListSelectionModel) e.getSource();

        if (m_lock || model.getValueIsAdjusting())
        {
            return;
        }

        if (FilamentSensor.DEBUG)
        {
            System.out.println(model);
        }

        int min = model.getMinSelectionIndex(), max = model.getMaxSelectionIndex();

        List<AbstractFilament> new_sel = new ArrayList<AbstractFilament>();

        for (int i = min; i <= max; i++)
        {
            if (model.isSelectedIndex(i))
            {
                new_sel.add(m_data.getEntryInRow(m_filament_table.convertRowIndexToModel(i)));
            }
        }
        m_data.setSelection(new_sel);
        m_image_panel.repaint();
    }

    /**
     * Handles PropertyChangeEvents sent by the SplitPane (dividerLocation).
     *
     * @param e
     *            The event object.(TODO: link?)
     */
    @Override
    public void propertyChange(PropertyChangeEvent e)
    {
        TabBar.tab_divider_location = m_picking_main_widget.getDividerLocation();
        if (FilamentSensor.DEBUG)
        {
            System.out.println("divider_location changed: " + TabBar.tab_divider_location + " (" + m_tab_title + ")");
        }
    }

}
